@@ -29,6 +29,7 @@ import dbus.mainloop.glib
 import slip.dbus
 
 from firewall.config import *
+from firewall.errors import FirewallError, ALREADY_ENABLED, NOT_ENABLED
 from firewall.config.dbus import *
 from firewall.core.base import DEFAULT_ZONE_TARGET
 from firewall.dbus_utils import dbus_to_python
@@ -37,6 +38,8 @@ from decorator import decorator
 from firewall.functions import b2u
 from firewall.core.rich import Rich_Rule
 
+import traceback
+
 exception_handler = None
 not_authorized_loop = False
 
@@ -56,19 +59,26 @@ def handle_exceptions(func, *args, **kwargs):
             if not exception_handler:
                 raise
             if "NotAuthorizedException" in dbus_name:
-                authorized = False
                 exception_handler("NotAuthorizedException")
+            elif "org.freedesktop.DBus.Error" in dbus_name:
+                # dbus error, try again
+                exception_handler(dbus_message)
             else:
                 authorized = True
                 if dbus_message:
                     exception_handler(dbus_message)
                 else:
                     exception_handler(b2u(str(e)))
-        except Exception as e:
+        except FirewallError as e:
             if not exception_handler:
                 raise
             else:
                 exception_handler(b2u(str(e)))
+        except Exception as e:
+            if not exception_handler:
+                raise
+            else:
+                exception_handler(b2u(traceback.format_exc()))
         if not not_authorized_loop:
             break
 
@@ -81,7 +91,7 @@ class FirewallClientZoneSettings(object):
             self.settings = settings
         else:
             self.settings = ["", "", "", False, DEFAULT_ZONE_TARGET, [], [],
-                             [], False, [], [], [], []]
+                             [], False, [], [], [], [], []]
 
     @handle_exceptions
     def __repr__(self):
@@ -127,10 +137,14 @@ class FirewallClientZoneSettings(object):
     def addService(self, service):
         if service not in self.settings[5]:
             self.settings[5].append(service)
+        else:
+            raise FirewallError(ALREADY_ENABLED, service)
     @handle_exceptions
     def removeService(self, service):
         if service in self.settings[5]:
             self.settings[5].remove(service)
+        else:
+            raise FirewallError(NOT_ENABLED, service)
     @handle_exceptions
     def queryService(self, service):
         return service in self.settings[5]
@@ -145,14 +159,40 @@ class FirewallClientZoneSettings(object):
     def addPort(self, port, protocol):
         if (port,protocol) not in self.settings[6]:
             self.settings[6].append((port,protocol))
+        else:
+            raise FirewallError(ALREADY_ENABLED, "'%s:%s'" % (port, protocol))
     @handle_exceptions
     def removePort(self, port, protocol):
         if (port,protocol) in self.settings[6]:
             self.settings[6].remove((port,protocol))
+        else:
+            raise FirewallError(NOT_ENABLED, "'%s:%s'" % (port, protocol))
     @handle_exceptions
     def queryPort(self, port, protocol):
         return (port,protocol) in self.settings[6]
 
+    @handle_exceptions
+    def getProtocols(self):
+        return self.settings[13]
+    @handle_exceptions
+    def setProtocols(self, protocols):
+        self.settings[13] = protocols
+    @handle_exceptions
+    def addProtocol(self, protocol):
+        if protocol not in self.settings[13]:
+            self.settings[13].append(protocol)
+        else:
+            raise FirewallError(ALREADY_ENABLED, protocol)
+    @handle_exceptions
+    def removeProtocol(self, protocol):
+        if protocol in self.settings[13]:
+            self.settings[13].remove(protocol)
+        else:
+            raise FirewallError(NOT_ENABLED, protocol)
+    @handle_exceptions
+    def queryProtocol(self, protocol):
+        return protocol in self.settings[13]
+
     @handle_exceptions
     def getIcmpBlocks(self):
         return sorted(self.settings[7])
@@ -163,10 +203,14 @@ class FirewallClientZoneSettings(object):
     def addIcmpBlock(self, icmptype):
         if icmptype not in self.settings[7]:
             self.settings[7].append(icmptype)
+        else:
+            raise FirewallError(ALREADY_ENABLED, icmptype)
     @handle_exceptions
     def removeIcmpBlock(self, icmptype):
         if icmptype in self.settings[7]:
             self.settings[7].remove(icmptype)
+        else:
+            raise FirewallError(NOT_ENABLED, icmptype)
     @handle_exceptions
     def queryIcmpBlock(self, icmptype):
         return icmptype in self.settings[7]
@@ -192,6 +236,9 @@ class FirewallClientZoneSettings(object):
             to_addr = ''
         if (port,protocol,to_port,to_addr) not in self.settings[9]:
             self.settings[9].append((port,protocol,to_port,to_addr))
+        else:
+            raise FirewallError(ALREADY_ENABLED, "'%s:%s:%s:%s'" % \
+                                (port, protocol, toport, toaddr))
     @handle_exceptions
     def removeForwardPort(self, port, protocol, to_port, to_addr):
         if to_port is None:
@@ -200,6 +247,9 @@ class FirewallClientZoneSettings(object):
             to_addr = ''
         if (port,protocol,to_port,to_addr) in self.settings[9]:
             self.settings[9].remove((port,protocol,to_port,to_addr))
+        else:
+            raise FirewallError(NOT_ENABLED, "'%s:%s:%s:%s'" % \
+                                (port, protocol, toport, toaddr))
     @handle_exceptions
     def queryForwardPort(self, port, protocol, to_port, to_addr):
         if to_port is None:
@@ -218,10 +268,14 @@ class FirewallClientZoneSettings(object):
     def addInterface(self, interface):
         if interface not in self.settings[10]:
             self.settings[10].append(interface)
+        else:
+            raise FirewallError(ALREADY_ENABLED, interface)
     @handle_exceptions
     def removeInterface(self, interface):
         if interface in self.settings[10]:
             self.settings[10].remove(interface)
+        else:
+            raise FirewallError(NOT_ENABLED, interface)
     @handle_exceptions
     def queryInterface(self, interface):
         return interface in self.settings[10]
@@ -236,10 +290,14 @@ class FirewallClientZoneSettings(object):
     def addSource(self, source):
         if source not in self.settings[11]:
             self.settings[11].append(source)
+        else:
+            raise FirewallError(ALREADY_ENABLED, source)
     @handle_exceptions
     def removeSource(self, source):
         if source in self.settings[11]:
             self.settings[11].remove(source)
+        else:
+            raise FirewallError(NOT_ENABLED, source)
     @handle_exceptions
     def querySource(self, source):
         return source in self.settings[11]
@@ -256,11 +314,15 @@ class FirewallClientZoneSettings(object):
         rule = str(Rich_Rule(rule_str=rule))
         if rule not in self.settings[12]:
             self.settings[12].append(rule)
+        else:
+            raise FirewallError(ALREADY_ENABLED, rule)
     @handle_exceptions
     def removeRichRule(self, rule):
         rule = str(Rich_Rule(rule_str=rule))
         if rule in self.settings[12]:
             self.settings[12].remove(rule)
+        else:
+            raise FirewallError(NOT_ENABLED, rule)
     @handle_exceptions
     def queryRichRule(self, rule):
         rule = str(Rich_Rule(rule_str=rule))
@@ -425,6 +487,33 @@ class FirewallClientConfigZone(object):
     def queryPort(self, port, protocol):
         return self.fw_zone.queryPort(port, protocol)
 
+    # protocol
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getProtocols(self):
+        return self.fw_zone.getProtocols()
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def setProtocols(self, protocols):
+        self.fw_zone.setProtocols(protocols)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def addProtocol(self, protocol):
+        self.fw_zone.addProtocol(protocol)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def removeProtocol(self, protocol):
+        self.fw_zone.removeProtocol(protocol)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def queryProtocol(self, protocol):
+        return self.fw_zone.queryProtocol(protocol)
+
     # icmp block
 
     @slip.dbus.polkit.enable_proxy
@@ -608,7 +697,7 @@ class FirewallClientServiceSettings(object):
         if settings:
             self.settings = settings
         else:
-            self.settings = ["", "", "", [], [], {}]
+            self.settings = ["", "", "", [], [], {}, []]
 
     @handle_exceptions
     def __repr__(self):
@@ -645,14 +734,40 @@ class FirewallClientServiceSettings(object):
     def addPort(self, port, protocol):
         if (port,protocol) not in self.settings[3]:
             self.settings[3].append((port,protocol))
+        else:
+            raise FirewallError(ALREADY_ENABLED, "'%s:%s'" % (port, protocol))
     @handle_exceptions
     def removePort(self, port, protocol):
         if (port,protocol) in self.settings[3]:
             self.settings[3].remove((port,protocol))
+        else:
+            raise FirewallError(NOT_ENABLED, "'%s:%s'" % (port, protocol))
     @handle_exceptions
     def queryPort(self, port, protocol):
         return (port,protocol) in self.settings[3]
 
+    @handle_exceptions
+    def getProtocols(self):
+        return self.settings[6]
+    @handle_exceptions
+    def setProtocols(self, protocols):
+        self.settings[6] = protocols
+    @handle_exceptions
+    def addProtocol(self, protocol):
+        if protocol not in self.settings[6]:
+            self.settings[6].append(protocol)
+        else:
+            raise FirewallError(ALREADY_ENABLED, protocol)
+    @handle_exceptions
+    def removeProtocol(self, protocol):
+        if protocol in self.settings[6]:
+            self.settings[6].remove(protocol)
+        else:
+            raise FirewallError(NOT_ENABLED, protocol)
+    @handle_exceptions
+    def queryProtocol(self, protocol):
+        return protocol in self.settings[6]
+
     @handle_exceptions
     def getModules(self):
         return sorted(self.settings[4])
@@ -663,10 +778,14 @@ class FirewallClientServiceSettings(object):
     def addModule(self, module):
         if module not in self.settings[4]:
             self.settings[4].append(module)
+        else:
+            raise FirewallError(ALREADY_ENABLED, module)
     @handle_exceptions
     def removeModule(self, module):
         if module in self.settings[4]:
             self.settings[4].remove(module)
+        else:
+            raise FirewallError(NOT_ENABLED, module)
     @handle_exceptions
     def queryModule(self, module):
         return module in self.settings[4]
@@ -679,16 +798,228 @@ class FirewallClientServiceSettings(object):
         self.settings[5] = destinations
     @handle_exceptions
     def setDestination(self, dest_type, address):
-        self.settings[5][dest_type] = address
+        if not dest_type in self.settings[5] or \
+           self.settings[5][dest_type] != address:
+            self.settings[5][dest_type] = address
+        else:
+            raise FirewallError(ALREADY_ENABLED, "'%s:%s'" % \
+                                (dest_type, address))
     @handle_exceptions
     def removeDestination(self, dest_type):
         if dest_type in self.settings[5]:
             del self.settings[5][dest_type]
+        else:
+            raise FirewallError(NOT_ENABLED, "'%s'" % dest_type)
     @handle_exceptions
     def queryDestination(self, dest_type, address):
         return (dest_type in self.settings[5] and \
                     address == self.settings[5][dest_type])
 
+# ipset config settings
+
+class FirewallClientIPSetSettings(object):
+    @handle_exceptions
+    def __init__(self, settings=None):
+        if settings:
+            self.settings = settings
+        else:
+            self.settings = ["", "", "", "", {}, []]
+
+    @handle_exceptions
+    def __repr__(self):
+        return '%s(%r)' % (self.__class__, self.settings)
+
+    @handle_exceptions
+    def getVersion(self):
+        return self.settings[0]
+    @handle_exceptions
+    def setVersion(self, version):
+        self.settings[0] = version
+
+    @handle_exceptions
+    def getShort(self):
+        return self.settings[1]
+    @handle_exceptions
+    def setShort(self, short):
+        self.settings[1] = short
+
+    @handle_exceptions
+    def getDescription(self):
+        return self.settings[2]
+    @handle_exceptions
+    def setDescription(self, description):
+        self.settings[2] = description
+
+    @handle_exceptions
+    def getType(self):
+        return self.settings[3]
+    @handle_exceptions
+    def setType(self, ipset_type):
+        self.settings[3] = ipset_type
+
+    @handle_exceptions
+    def getOptions(self):
+        return self.settings[4]
+    @handle_exceptions
+    def setOptions(self, options):
+        self.settings[4] = options
+    @handle_exceptions
+    def addOption(self, key, value):
+        if not key in self.settings[4] or self.settings[4][key] != value:
+            self.settings[4][key] = value
+        else:
+            raise FirewallError(ALREADY_ENABLED, "'%s=%s'" % (key,value)
+                                if value else key)
+    @handle_exceptions
+    def removeOption(self, key):
+        if key in self.settings[4]:
+            del self.settings[4][key]
+        else:
+            raise FirewallError(NOT_ENABLED, key)
+    @handle_exceptions
+    def queryOption(self, key, value):
+        return key in self.settings[4] and self.settings[4][key] == value
+
+    @handle_exceptions
+    def getEntries(self):
+        return self.settings[5]
+    @handle_exceptions
+    def setEntries(self, entries):
+        self.settings[5] = entries
+    @handle_exceptions
+    def addEntry(self, entry):
+        if entry not in self.settings[5]:
+            self.settings[5].append(entry)
+        else:
+            raise FirewallError(ALREADY_ENABLED, entry)
+    @handle_exceptions
+    def removeEntry(self, entry):
+        if entry in self.settings[5]:
+            self.settings[5].remove(entry)
+        else:
+            raise FirewallError(NOT_ENABLED, entry)
+    @handle_exceptions
+    def queryEntry(self, entry):
+        return entry in self.settings[5]
+
+# ipset config
+
+class FirewallClientConfigIPSet(object):
+    @handle_exceptions
+    def __init__(self, bus, path):
+        self.bus = bus
+        self.path = path
+        self.dbus_obj = self.bus.get_object(DBUS_INTERFACE, path)
+        self.fw_ipset = dbus.Interface(self.dbus_obj,
+                                       dbus_interface=DBUS_INTERFACE_CONFIG_IPSET)
+        self.fw_properties = dbus.Interface(
+            self.dbus_obj, dbus_interface='org.freedesktop.DBus.Properties')
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def get_property(self, prop):
+        return dbus_to_python(self.fw_properties.Get(DBUS_INTERFACE_CONFIG_IPSET,
+                                                     prop))
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def get_properties(self):
+        return dbus_to_python(self.fw_properties.GetAll(DBUS_INTERFACE_CONFIG_IPSET))
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def set_property(self, prop, value):
+        self.fw_properties.Set(DBUS_INTERFACE_CONFIG_IPSET, prop, value)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getSettings(self):
+        return FirewallClientIPSetSettings(list(dbus_to_python(\
+                    self.fw_ipset.getSettings())))
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def update(self, settings):
+        self.fw_ipset.update(tuple(settings.settings))
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def loadDefaults(self):
+        self.fw_ipset.loadDefaults()
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def remove(self):
+        self.fw_ipset.remove()
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def rename(self, name):
+        self.fw_ipset.rename(name)
+
+    # version
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getVersion(self):
+        return self.fw_ipset.getVersion()
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def setVersion(self, version):
+        self.fw_ipset.setVersion(version)
+
+    # short
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getShort(self):
+        return self.fw_ipset.getShort()
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def setShort(self, short):
+        self.fw_ipset.setShort(short)
+
+    # description
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getDescription(self):
+        return self.fw_ipset.getDescription()
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def setDescription(self, description):
+        self.fw_ipset.setDescription(description)
+
+    # entry
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getEntries(self):
+        return self.fw_ipset.getEntries()
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def setEntries(self, entries):
+        self.fw_ipset.setEntries(entries)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def addEntry(self, entry):
+        self.fw_ipset.addEntry(entry)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def removeEntry(self, entry):
+        self.fw_ipset.removeEntry(entry)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def queryEntry(self, entry):
+        return self.fw_ipset.queryEntry(entry)
+
 # service config
 
 class FirewallClientConfigService(object):
@@ -807,6 +1138,33 @@ class FirewallClientConfigService(object):
     def queryPort(self, port, protocol):
         return self.fw_service.queryPort(port, protocol)
 
+    # protocol
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getProtocols(self):
+        return self.fw_service.getProtocols()
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def setProtocols(self, protocols):
+        self.fw_service.setProtocols(protocols)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def addProtocol(self, protocol):
+        self.fw_service.addProtocol(protocol)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def removeProtocol(self, protocol):
+        self.fw_service.removeProtocol(protocol)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def queryProtocol(self, protocol):
+        return self.fw_service.queryProtocol(protocol)
+
     # module
 
     @slip.dbus.polkit.enable_proxy
@@ -858,8 +1216,8 @@ class FirewallClientConfigService(object):
 
     @slip.dbus.polkit.enable_proxy
     @handle_exceptions
-    def removeDestination(self, destination, address):
-        self.fw_service.removeDestination(destination, address)
+    def removeDestination(self, destination):
+        self.fw_service.removeDestination(destination)
 
     @slip.dbus.polkit.enable_proxy
     @handle_exceptions
@@ -1487,6 +1845,30 @@ class FirewallClientConfig(object):
     def set_property(self, prop, value):
         self.fw_properties.Set(DBUS_INTERFACE_CONFIG, prop, value)
 
+    # ipset
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def listIPSets(self):
+        return dbus_to_python(self.fw_config.listIPSets())
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getIPSet(self, path):
+        return FirewallClientConfigIPSet(self.bus, path)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getIPSetByName(self, name):
+        path = dbus_to_python(self.fw_config.getIPSetByName(name))
+        return FirewallClientConfigIPSet(self.bus, path)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def addIPSet(self, name, settings):
+        path = self.fw_config.addIPSet(name, tuple(settings.settings))
+        return FirewallClientConfigIPSet(self.bus, path)
+
     # zone
 
     @slip.dbus.polkit.enable_proxy
@@ -1602,10 +1984,12 @@ class FirewallClient(object):
             arg0=DBUS_INTERFACE)
 
         for interface in [ DBUS_INTERFACE,
+                           DBUS_INTERFACE_IPSET,
                            DBUS_INTERFACE_ZONE,
                            DBUS_INTERFACE_DIRECT,
                            DBUS_INTERFACE_POLICIES,
                            DBUS_INTERFACE_CONFIG,
+                           DBUS_INTERFACE_CONFIG_IPSET,
                            DBUS_INTERFACE_CONFIG_ZONE,
                            DBUS_INTERFACE_CONFIG_SERVICE,
                            DBUS_INTERFACE_CONFIG_DIRECT,
@@ -1625,6 +2009,7 @@ class FirewallClient(object):
             "connection-established": "connection-established",
             "connection-lost": "connection-lost",
             # firewalld callbacks
+            "log-denied-changed": "LogDeniedChanged",
             "default-zone-changed": "DefaultZoneChanged",
             "panic-mode-enabled": "PanicModeEnabled",
             "panic-mode-disabled": "PanicModeDisabled",
@@ -1633,6 +2018,8 @@ class FirewallClient(object):
             "service-removed": "ServiceRemoved",
             "port-added": "PortAdded",
             "port-removed": "PortRemoved",
+            "protocol-added": "ProtocolAdded",
+            "protocol-removed": "ProtocolRemoved",
             "masquerade-added": "MasqueradeAdded",
             "masquerade-removed": "MasqueradeRemoved",
             "forward-port-added": "ForwardPortAdded",
@@ -1648,6 +2035,9 @@ class FirewallClient(object):
             "source-added": "SourceAdded",
             "source-removed": "SourceRemoved",
             "zone-of-source-changed": "ZoneOfSourceChanged",
+            # ipset callbacks
+            "ipset-entry-added": "EntryAdded",
+            "ipset-entry-removed": "EntryRemoved",
             # direct callbacks
             "direct:chain-added": "ChainAdded",
             "direct:chain-removed": "ChainRemoved",
@@ -1669,6 +2059,10 @@ class FirewallClient(object):
             "lockdown-whitelist-user-removed": "LockdownWhitelistUserRemoved",
             # firewalld.config callbacks
             "config:policies:lockdown-whitelist-updated": "config:policies:LockdownWhitelistUpdated",
+            "config:ipset-added": "config:IPSetAdded",
+            "config:ipset-updated": "config:IPSetUpdated",
+            "config:ipset-removed": "config:IPSetRemoved",
+            "config:ipset-renamed": "config:IPSetRenamed",
             "config:zone-added": "config:ZoneAdded",
             "config:zone-updated": "config:ZoneUpdated",
             "config:zone-removed": "config:ZoneRemoved",
@@ -1697,6 +2091,7 @@ class FirewallClient(object):
     @handle_exceptions
     def _init_vars(self):
         self.fw = None
+        self.fw_ipset = None
         self.fw_zone = None
         self.fw_direct = None
         self.fw_properties = None
@@ -1748,6 +2143,8 @@ class FirewallClient(object):
             self.dbus_obj = self.bus.get_object(DBUS_INTERFACE, DBUS_PATH)
             self.fw = dbus.Interface(self.dbus_obj,
                                      dbus_interface=DBUS_INTERFACE)
+            self.fw_ipset = dbus.Interface(self.dbus_obj,
+                                           dbus_interface=DBUS_INTERFACE_IPSET)
             self.fw_zone = dbus.Interface(self.dbus_obj,
                                           dbus_interface=DBUS_INTERFACE_ZONE)
             self.fw_direct = dbus.Interface(
@@ -1798,6 +2195,8 @@ class FirewallClient(object):
         # pimp signal name
         if interface.startswith(DBUS_INTERFACE_CONFIG_ZONE):
             signal = "config:Zone" + signal
+        elif interface.startswith(DBUS_INTERFACE_CONFIG_IPSET):
+            signal = "config:IPSet" + signal
         elif interface.startswith(DBUS_INTERFACE_CONFIG_SERVICE):
             signal = "config:Service" + signal
         elif interface.startswith(DBUS_INTERFACE_CONFIG_ICMPTYPE):
@@ -1888,6 +2287,37 @@ class FirewallClient(object):
         return FirewallClientZoneSettings(list(dbus_to_python(\
                     self.fw.getZoneSettings(zone))))
 
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getIPSets(self):
+        return dbus_to_python(self.fw_ipset.getIPSets())
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getIPSetSettings(self, ipset):
+        return FirewallClientIPSetSettings(list(dbus_to_python(\
+                    self.fw_ipset.getIPSetSettings(ipset))))
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def addEntry(self, ipset, entry):
+        self.fw_ipset.addEntry(ipset, entry)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getEntries(self, ipset):
+        return self.fw_ipset.getEntries(ipset)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def removeEntry(self, ipset, entry):
+        self.fw_ipset.removeEntry(ipset, entry)
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def queryEntry(self, ipset, entry):
+        return dbus_to_python(self.fw_ipset.queryEntry(ipset, entry))
+
     @slip.dbus.polkit.enable_proxy
     @handle_exceptions
     def listServices(self):
@@ -1910,6 +2340,18 @@ class FirewallClient(object):
         return FirewallClientIcmpTypeSettings(list(dbus_to_python(\
                     self.fw.getIcmpTypeSettings(icmptype))))
 
+    # log denied
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getLogDenied(self):
+        return dbus_to_python(self.fw.getLogDenied())
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def setLogDenied(self, value):
+        self.fw.setLogDenied(value)
+
     # default zone
 
     @slip.dbus.polkit.enable_proxy
@@ -2075,6 +2517,28 @@ class FirewallClient(object):
     def removePort(self, zone, port, protocol):
         return dbus_to_python(self.fw_zone.removePort(zone, port, protocol))
 
+    # protocols
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def addProtocol(self, zone, protocol, timeout=0):
+        return dbus_to_python(self.fw_zone.addProtocol(zone, protocol, timeout))
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def getProtocols(self, zone):
+        return dbus_to_python(self.fw_zone.getProtocols(zone))
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def queryProtocol(self, zone, protocol):
+        return dbus_to_python(self.fw_zone.queryProtocol(zone, protocol))
+
+    @slip.dbus.polkit.enable_proxy
+    @handle_exceptions
+    def removeProtocol(self, zone, protocol):
+        return dbus_to_python(self.fw_zone.removeProtocol(zone, protocol))
+
     # masquerade
 
     @slip.dbus.polkit.enable_proxy
