@@ -30,27 +30,3 @@ SHORTCUTS = {
     "FORWARD_OUT": "FWDO",
     "OUTPUT": "OUT",
 }
-
-INTERFACE_ZONE_OPTS = {
-    "PREROUTING": "-i",
-    "POSTROUTING": "-o",
-    "INPUT": "-i",
-    "FORWARD_IN": "-i",
-    "FORWARD_OUT": "-o",
-    "OUTPUT": "-o",
-}
-
-ZONE_CHAINS = {
-    "filter": {
-        "INPUT": [ "ipv4", "ipv6" ],
-        "FORWARD_IN": [ "ipv4", "ipv6" ],
-        "FORWARD_OUT": [ "ipv4", "ipv6" ],
-        },
-    "nat": {
-        "PREROUTING": [ "ipv4", "ipv6" ],
-        "POSTROUTING": [ "ipv4", "ipv6" ],
-        },
-    "mangle": {
-        "PREROUTING": [ "ipv4", "ipv6" ],
-        },
-}
