@@ -27,49 +27,59 @@ from firewall.core.logger import log
 from firewall.core.io.icmptype import IcmpType, icmptype_reader, icmptype_writer
 from firewall.core.io.service import Service, service_reader, service_writer
 from firewall.core.io.zone import Zone, zone_reader, zone_writer
+from firewall.core.io.ipset import IPSet, ipset_reader, ipset_writer
 from firewall.functions import portStr
 from firewall.errors import *
 
-class FirewallConfig:
+class FirewallConfig(object):
     def __init__(self, fw):
         self._fw = fw
         self.__init_vars()
 
     def __repr__(self):
-        return '%s(%r, %r, %r, %r, %r, %r, %r, %r, %r)' % (self.__class__,
-                   self._icmptypes, self._services, self._zones,
-                   self._default_icmptypes, self._default_services, self._default_zones,
+        return '%s(%r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r)' % (self.__class__,
+                   self._ipsets, self._icmptypes, self._services, self._zones,
+                   self._builtin_ipsets, self._builtin_icmptypes, self._builtin_services, self._builtin_zones,
                    self._firewalld_conf, self._policies, self._direct)
 
     def __init_vars(self):
+        self._ipsets = { }
         self._icmptypes = { }
         self._services = { }
         self._zones = { }
-        self._default_icmptypes = { }
-        self._default_services = { }
-        self._default_zones = { }
+        self._builtin_ipsets = { }
+        self._builtin_icmptypes = { }
+        self._builtin_services = { }
+        self._builtin_zones = { }
         self._firewalld_conf = None
         self._policies = None
         self._direct = None
 
     def cleanup(self):
-        for x in list(self._default_icmptypes.keys()):
-            self._default_icmptypes[x].cleanup()
-            del self._default_icmptypes[x]
+        for x in list(self._builtin_ipsets.keys()):
+            self._builtin_ipsets[x].cleanup()
+            del self._builtin_ipsets[x]
+        for x in list(self._ipsets.keys()):
+            self._ipsets[x].cleanup()
+            del self._ipsets[x]
+
+        for x in list(self._builtin_icmptypes.keys()):
+            self._builtin_icmptypes[x].cleanup()
+            del self._builtin_icmptypes[x]
         for x in list(self._icmptypes.keys()):
             self._icmptypes[x].cleanup()
             del self._icmptypes[x]
 
-        for x in list(self._default_services.keys()):
-            self._default_services[x].cleanup()
-            del self._default_services[x]
+        for x in list(self._builtin_services.keys()):
+            self._builtin_services[x].cleanup()
+            del self._builtin_services[x]
         for x in list(self._services.keys()):
             self._services[x].cleanup()
             del self._services[x]
 
-        for x in list(self._default_zones.keys()):
-            self._default_zones[x].cleanup()
-            del self._default_zones[x]
+        for x in list(self._builtin_zones.keys()):
+            self._builtin_zones[x].cleanup()
+            del self._builtin_zones[x]
         for x in list(self._zones.keys()):
             self._zones[x].cleanup()
             del self._zones[x]
@@ -108,7 +118,7 @@ class FirewallConfig:
         return self._firewalld_conf
 
     def update_firewalld_conf(self):
-        if not os.path.exists(LOCKDOWN_WHITELIST):
+        if not os.path.exists(FIREWALLD_CONF):
             self._firewalld_conf.clear()
         else:
             self._firewalld_conf.read()
@@ -141,53 +151,216 @@ class FirewallConfig:
         else:
             self._direct.read()
 
+    # ipset
+
+    def get_ipsets(self):
+        return sorted(set(list(self._ipsets.keys()) + \
+                          list(self._builtin_ipsets.keys())))
+
+    def add_ipset(self, obj):
+        if obj.builtin:
+            self._builtin_ipsets[obj.name] = obj
+        else:
+            self._ipsets[obj.name] = obj
+
+    def get_ipset(self, name):
+        if name in self._ipsets:
+            return self._ipsets[name]
+        elif name in self._builtin_ipsets:
+            return self._builtin_ipsets[name]
+        raise FirewallError(INVALID_IPSET, name)
+
+    def load_ipset_defaults(self, obj):
+        if obj.name not in self._ipsets:
+            raise FirewallError(NO_DEFAULTS, obj.name)
+        elif self._ipsets[obj.name] != obj:
+            raise FirewallError(NO_DEFAULTS,
+                                "self._ipsets[%s] != obj" % obj.name)
+        elif obj.name not in self._builtin_ipsets:
+            raise FirewallError(NO_DEFAULTS,
+                            "'%s' not in self._builtin_ipsets" % obj.name)
+        self._remove_ipset(obj)
+        return self._builtin_ipsets[obj.name]
+
+    def get_ipset_config(self, obj):
+        return obj.export_config()
+
+    def set_ipset_config(self, obj, config):
+        if obj.builtin:
+            x = copy.copy(obj)
+            x.import_config(config)
+            x.path = ETC_FIREWALLD_IPSETS
+            x.builtin = False
+            if obj.path != x.path:
+                x.default = False
+            self.add_ipset(x)
+            ipset_writer(x)
+            return x
+        else:
+            obj.import_config(config)
+            ipset_writer(obj)
+            return obj
+
+    def new_ipset(self, name, config):
+        try:
+            self.get_ipset(name)
+        except:
+            pass
+        else:
+            raise FirewallError(NAME_CONFLICT, "new_ipset(): '%s'" % name)
+
+        x = IPSet()
+        x.check_name(name)
+        x.import_config(config)
+        x.name = name
+        x.filename = "%s.xml" % name
+        x.path = ETC_FIREWALLD_IPSETS
+        # It is not possible to add a new one with a name of a buitin
+        x.builtin = False
+        x.default = True
+
+        ipset_writer(x)
+        self.add_ipset(x)
+        return x
+
+    def update_ipset_from_path(self, name):
+        filename = os.path.basename(name)
+        path = os.path.dirname(name)
+
+        if not os.path.exists(name):
+            # removed file
+
+            if path == ETC_FIREWALLD_IPSETS:
+                # removed custom ipset
+                for x in self._ipsets.keys():
+                    obj = self._ipsets[x]
+                    if obj.filename == filename:
+                        del self._ipsets[x]
+                        if obj.name in self._builtin_ipsets:
+                            return ("update", self._builtin_ipsets[obj.name])
+                        return ("remove", obj)
+            else:
+                # removed builtin ipset
+                for x in self._builtin_ipsets.keys():
+                    obj = self._builtin_ipsets[x]
+                    if obj.filename == filename:
+                        del self._builtin_ipsets[x]
+                        if obj.name not in self._ipsets:
+                            # update dbus ipset
+                            return ("remove", obj)
+                        else:
+                            # builtin hidden, no update needed
+                            return (None, None)
+
+            # ipset not known to firewalld, yet (timeout, ..)
+            return (None, None)
+
+        # new or updated file
+
+        log.debug1("Loading ipset file '%s'", name)
+        try:
+            obj = ipset_reader(filename, path)
+        except Exception as msg:
+            log.error("Failed to load ipset file '%s': %s", filename, msg)
+            return (None, None)
+
+        # new ipset
+        if obj.name not in self._builtin_ipsets and obj.name not in self._ipsets:
+            self.add_ipset(obj)
+            return ("new", obj)
+
+        # updated ipset
+        if path == ETC_FIREWALLD_IPSETS:
+            # custom ipset update
+            if obj.name in self._ipsets:
+                self._ipsets[obj.name] = obj
+            return ("update", obj)
+        else:
+            if obj.name in self._builtin_ipsets:
+                # builtin ipset update
+                del self._builtin_ipsets[obj.name]
+                self._builtin_ipsets[obj.name] = obj
+
+                if obj.name not in self._ipsets:
+                    # update dbus ipset
+                    return ("update", obj)
+                else:
+                    # builtin hidden, no update needed
+                    return (None, None)
+
+        # ipset not known to firewalld, yet (timeout, ..)
+        return (None, None)
+
+    def _remove_ipset(self, obj):
+        if obj.name not in self._ipsets:
+            raise FirewallError(INVALID_IPSET,
+                                "'%s' not in self._ipsets" % obj.name)
+        if obj.path != ETC_FIREWALLD_IPSETS:
+            raise FirewallError(INVALID_DIRECTORY,
+                        "'%s' != '%s'" % (obj.path, ETC_FIREWALLD_IPSETS))
+        os.remove("%s/%s.xml" % (obj.path, obj.name))
+        del self._ipsets[obj.name]
+
+    def check_builtin_ipset(self, obj):
+        if obj.builtin:
+            raise FirewallError(BUILTIN_IPSET,
+                                "'%s' is built-in icmp type" % obj.name)
+
+    def remove_ipset(self, obj):
+        self.check_builtin_ipset(obj)
+        self._remove_ipset(obj)
+
+    def rename_ipset(self, obj, name):
+        self.check_builtin_ipset(obj)
+        new_ipset = self._copy_ipset(obj, name)
+        self._remove_ipset(obj)
+        return new_ipset
+
+    def _copy_ipset(self, obj, name):
+        return self.new_ipset(name, obj.export_config())
+
     # icmptypes
 
     def get_icmptypes(self):
         return sorted(set(list(self._icmptypes.keys()) + \
-                          list(self._default_icmptypes.keys())))
+                          list(self._builtin_icmptypes.keys())))
 
     def add_icmptype(self, obj):
-        if obj.default:
-            self._default_icmptypes[obj.name] = obj
+        if obj.builtin:
+            self._builtin_icmptypes[obj.name] = obj
         else:
             self._icmptypes[obj.name] = obj
 
     def get_icmptype(self, name):
         if name in self._icmptypes:
             return self._icmptypes[name]
-        elif name in self._default_icmptypes:
-            return self._default_icmptypes[name]
+        elif name in self._builtin_icmptypes:
+            return self._builtin_icmptypes[name]
         raise FirewallError(INVALID_ICMPTYPE, name)
 
-    def icmptype_has_defaults(self, name):
-        return (name in self._icmptypes and name in self._default_icmptypes)
-
-    def icmptype_is_default(self, name):
-        return name in self._default_icmptypes
-
     def load_icmptype_defaults(self, obj):
         if obj.name not in self._icmptypes:
-            raise FirewallError(NO_DEFAULTS,
-                                "'%s' not in self._icmptypes" % obj.name)
+            raise FirewallError(NO_DEFAULTS, obj.name)
         elif self._icmptypes[obj.name] != obj:
             raise FirewallError(NO_DEFAULTS,
                                 "self._icmptypes[%s] != obj" % obj.name)
-        elif obj.name not in self._default_icmptypes:
+        elif obj.name not in self._builtin_icmptypes:
             raise FirewallError(NO_DEFAULTS,
-                            "'%s' not in self._default_icmptypes" % obj.name)
+                            "'%s' not in self._builtin_icmptypes" % obj.name)
         self._remove_icmptype(obj)
-        return self._default_icmptypes[obj.name]
+        return self._builtin_icmptypes[obj.name]
 
     def get_icmptype_config(self, obj):
         return obj.export_config()
 
     def set_icmptype_config(self, obj, config):
-        if obj.default:
+        if obj.builtin:
             x = copy.copy(obj)
             x.import_config(config)
             x.path = ETC_FIREWALLD_ICMPTYPES
-            x.default = False
+            x.builtin = False
+            if obj.path != x.path:
+                x.default = False
             self.add_icmptype(x)
             icmptype_writer(x)
             return x
@@ -210,7 +383,9 @@ class FirewallConfig:
         x.name = name
         x.filename = "%s.xml" % name
         x.path = ETC_FIREWALLD_ICMPTYPES
-        x.default = False
+        # It is not possible to add a new one with a name of a buitin
+        x.builtin = False
+        x.default = True
 
         icmptype_writer(x)
         self.add_icmptype(x)
@@ -229,15 +404,15 @@ class FirewallConfig:
                     obj = self._icmptypes[x]
                     if obj.filename == filename:
                         del self._icmptypes[x]
-                        if obj.name in self._default_icmptypes:
-                            return ("update", self._default_icmptypes[obj.name])
+                        if obj.name in self._builtin_icmptypes:
+                            return ("update", self._builtin_icmptypes[obj.name])
                         return ("remove", obj)
             else:
                 # removed builtin icmptype
-                for x in self._default_icmptypes.keys():
-                    obj = self._default_icmptypes[x]
+                for x in self._builtin_icmptypes.keys():
+                    obj = self._builtin_icmptypes[x]
                     if obj.filename == filename:
-                        del self._default_icmptypes[x]
+                        del self._builtin_icmptypes[x]
                         if obj.name not in self._icmptypes:
                             # update dbus icmptype
                             return ("remove", obj)
@@ -250,10 +425,15 @@ class FirewallConfig:
 
         # new or updated file
 
-        obj = icmptype_reader(filename, path)
+        log.debug1("Loading icmptype file '%s'", name)
+        try:
+            obj = icmptype_reader(filename, path)
+        except Exception as msg:
+            log.error("Failed to load icmptype file '%s': %s", filename, msg)
+            return (None, None)
 
         # new icmptype
-        if obj.name not in self._default_icmptypes and obj.name not in self._icmptypes:
+        if obj.name not in self._builtin_icmptypes and obj.name not in self._icmptypes:
             self.add_icmptype(obj)
             return ("new", obj)
 
@@ -264,10 +444,10 @@ class FirewallConfig:
                 self._icmptypes[obj.name] = obj
             return ("update", obj)
         else:
-            if obj.name in self._default_icmptypes:
+            if obj.name in self._builtin_icmptypes:
                 # builtin icmptype update
-                del self._default_icmptypes[obj.name]
-                self._default_icmptypes[obj.name] = obj
+                del self._builtin_icmptypes[obj.name]
+                self._builtin_icmptypes[obj.name] = obj
 
                 if obj.name not in self._icmptypes:
                     # update dbus icmptype
@@ -289,13 +469,8 @@ class FirewallConfig:
         os.remove("%s/%s.xml" % (obj.path, obj.name))
         del self._icmptypes[obj.name]
 
-    def is_builtin_icmptype(self, obj):
-        if obj.default or obj.name in self._default_icmptypes:
-            return True
-        return False
-
     def check_builtin_icmptype(self, obj):
-        if self.is_builtin_icmptype(obj):
+        if obj.builtin:
             raise FirewallError(BUILTIN_ICMPTYPE,
                                 "'%s' is built-in icmp type" % obj.name)
 
@@ -316,49 +491,44 @@ class FirewallConfig:
 
     def get_services(self):
         return sorted(set(list(self._services.keys()) + \
-                          list(self._default_services.keys())))
+                          list(self._builtin_services.keys())))
 
     def add_service(self, obj):
-        if obj.default:
-            self._default_services[obj.name] = obj
+        if obj.builtin:
+            self._builtin_services[obj.name] = obj
         else:
             self._services[obj.name] = obj
 
     def get_service(self, name):
         if name in self._services:
             return self._services[name]
-        elif name in self._default_services:
-            return self._default_services[name]
+        elif name in self._builtin_services:
+            return self._builtin_services[name]
         raise FirewallError(INVALID_SERVICE, "get_service(): '%s'" % name)
 
-    def service_has_defaults(self, name):
-        return (name in self._services and name in self._default_services)
-
-    def service_is_default(self, name):
-        return name in self._default_services
-
     def load_service_defaults(self, obj):
         if obj.name not in self._services:
-            raise FirewallError(NO_DEFAULTS,
-                                "'%s' not in self._services" % obj.name)
+            raise FirewallError(NO_DEFAULTS, obj.name)
         elif self._services[obj.name] != obj:
             raise FirewallError(NO_DEFAULTS,
                                 "self._services[%s] != obj" % obj.name)
-        elif obj.name not in self._default_services:
+        elif obj.name not in self._builtin_services:
             raise FirewallError(NO_DEFAULTS,
-                            "'%s' not in self._default_services" % obj.name)
+                            "'%s' not in self._builtin_services" % obj.name)
         self._remove_service(obj)
-        return self._default_services[obj.name]
+        return self._builtin_services[obj.name]
 
     def get_service_config(self, obj):
         return obj.export_config()
 
     def set_service_config(self, obj, config):
-        if obj.default:
+        if obj.builtin:
             x = copy.copy(obj)
             x.import_config(config)
             x.path = ETC_FIREWALLD_SERVICES
-            x.default = False
+            x.builtin = False
+            if obj.path != x.path:
+                x.default = False
             self.add_service(x)
             service_writer(x)
             return x
@@ -381,7 +551,9 @@ class FirewallConfig:
         x.name = name
         x.filename = "%s.xml" % name
         x.path = ETC_FIREWALLD_SERVICES
-        x.default = False
+        # It is not possible to add a new one with a name of a buitin
+        x.builtin = False
+        x.default = True
 
         service_writer(x)
         self.add_service(x)
@@ -400,15 +572,15 @@ class FirewallConfig:
                     obj = self._services[x]
                     if obj.filename == filename:
                         del self._services[x]
-                        if obj.name in self._default_services:
-                            return ("update", self._default_services[obj.name])
+                        if obj.name in self._builtin_services:
+                            return ("update", self._builtin_services[obj.name])
                         return ("remove", obj)
             else:
                 # removed builtin service
-                for x in self._default_services.keys():
-                    obj = self._default_services[x]
+                for x in self._builtin_services.keys():
+                    obj = self._builtin_services[x]
                     if obj.filename == filename:
-                        del self._default_services[x]
+                        del self._builtin_services[x]
                         if obj.name not in self._services:
                             # update dbus service
                             return ("remove", obj)
@@ -421,10 +593,15 @@ class FirewallConfig:
 
         # new or updated file
 
-        obj = service_reader(filename, path)
+        log.debug1("Loading service file '%s'", name)
+        try:
+            obj = service_reader(filename, path)
+        except Exception as msg:
+            log.error("Failed to load service file '%s': %s", filename, msg)
+            return (None, None)
 
         # new service
-        if obj.name not in self._default_services and obj.name not in self._services:
+        if obj.name not in self._builtin_services and obj.name not in self._services:
             self.add_service(obj)
             return ("new", obj)
 
@@ -435,10 +612,10 @@ class FirewallConfig:
                 self._services[obj.name] = obj
             return ("update", obj)
         else:
-            if obj.name in self._default_services:
+            if obj.name in self._builtin_services:
                 # builtin service update
-                del self._default_services[obj.name]
-                self._default_services[obj.name] = obj
+                del self._builtin_services[obj.name]
+                self._builtin_services[obj.name] = obj
 
                 if obj.name not in self._services:
                     # update dbus service
@@ -460,13 +637,8 @@ class FirewallConfig:
         os.remove("%s/%s.xml" % (obj.path, obj.name))
         del self._services[obj.name]
 
-    def is_builtin_service(self, obj):
-        if obj.default or obj.name in self._default_services:
-            return True
-        return False
-
     def check_builtin_service(self, obj):
-        if self.is_builtin_service(obj):
+        if obj.builtin:
             raise FirewallError(BUILTIN_SERVICE,
                                 "'%s' is built-in service" % obj.name)
 
@@ -487,56 +659,51 @@ class FirewallConfig:
 
     def get_zones(self):
         return sorted(set(list(self._zones.keys()) + \
-                          list(self._default_zones.keys())))
+                          list(self._builtin_zones.keys())))
 
     def add_zone(self, obj):
-        if obj.default:
-            self._default_zones[obj.name] = obj
+        if obj.builtin:
+            self._builtin_zones[obj.name] = obj
         else:
             self._zones[obj.name] = obj
 
     def forget_zone(self, name):
-        if name in self._default_zones:
-            del self._default_zones[name]
+        if name in self._builtin_zones:
+            del self._builtin_zones[name]
         if name in self._zones:
             del self._zones[name]
 
     def get_zone(self, name):
         if name in self._zones:
             return self._zones[name]
-        elif name in self._default_zones:
-            return self._default_zones[name]
+        elif name in self._builtin_zones:
+            return self._builtin_zones[name]
         raise FirewallError(INVALID_ZONE, "get_zone(): %s" % name)
 
-    def zone_has_defaults(self, name):
-        return (name in self._zones and name in self._default_zones)
-
-    def zone_is_default(self, name):
-        return name in self._default_zones
-
     def load_zone_defaults(self, obj):
         if obj.name not in self._zones:
-            raise FirewallError(NO_DEFAULTS,
-                                "'%s' not in self._zones" % obj.name)
+            raise FirewallError(NO_DEFAULTS, obj.name)
         elif self._zones[obj.name] != obj:
             raise FirewallError(NO_DEFAULTS,
                                 "self._zones[%s] != obj" % obj.name)
-        elif obj.name not in self._default_zones:
+        elif obj.name not in self._builtin_zones:
             raise FirewallError(NO_DEFAULTS,
-                                "'%s' not in self._default_zones" % obj.name)
+                                "'%s' not in self._builtin_zones" % obj.name)
         self._remove_zone(obj)
-        return self._default_zones[obj.name]
+        return self._builtin_zones[obj.name]
 
     def get_zone_config(self, obj):
         return obj.export_config()
 
     def set_zone_config(self, obj, config):
-        if obj.default:
+        if obj.builtin:
             x = copy.copy(obj)
             x.fw_config = self
             x.import_config(config)
             x.path = ETC_FIREWALLD_ZONES
-            x.default = False
+            x.builtin = False
+            if obj.path != x.path:
+                x.default = False
             self.add_zone(x)
             zone_writer(x)
             return x
@@ -561,7 +728,9 @@ class FirewallConfig:
         x.name = name
         x.filename = "%s.xml" % name
         x.path = ETC_FIREWALLD_ZONES
-        x.default = False
+        # It is not possible to add a new one with a name of a buitin
+        x.builtin = False
+        x.default = True
 
         zone_writer(x)
         self.add_zone(x)
@@ -580,15 +749,15 @@ class FirewallConfig:
                     obj = self._zones[x]
                     if obj.filename == filename:
                         del self._zones[x]
-                        if obj.name in self._default_zones:
-                            return ("update", self._default_zones[obj.name])
+                        if obj.name in self._builtin_zones:
+                            return ("update", self._builtin_zones[obj.name])
                         return ("remove", obj)
             else:
                 # removed builtin zone
-                for x in self._default_zones.keys():
-                    obj = self._default_zones[x]
+                for x in self._builtin_zones.keys():
+                    obj = self._builtin_zones[x]
                     if obj.filename == filename:
-                        del self._default_zones[x]
+                        del self._builtin_zones[x]
                         if obj.name not in self._zones:
                             # update dbus zone
                             return ("remove", obj)
@@ -601,11 +770,17 @@ class FirewallConfig:
 
         # new or updated file
 
-        obj = zone_reader(filename, path)
+        log.debug1("Loading zone file '%s'", name)
+        try:
+            obj = zone_reader(filename, path)
+        except Exception as msg:
+            log.error("Failed to load zone file '%s': %s", filename, msg)
+            return (None, None)
+
         obj.fw_config = self
 
         # new zone
-        if obj.name not in self._default_zones and obj.name not in self._zones:
+        if obj.name not in self._builtin_zones and obj.name not in self._zones:
             self.add_zone(obj)
             return ("new", obj)
 
@@ -616,10 +791,10 @@ class FirewallConfig:
                 self._zones[obj.name] = obj
             return ("update", obj)
         else:
-            if obj.name in self._default_zones:
+            if obj.name in self._builtin_zones:
                 # builtin zone update
-                del self._default_zones[obj.name]
-                self._default_zones[obj.name] = obj
+                del self._builtin_zones[obj.name]
+                self._builtin_zones[obj.name] = obj
 
                 if obj.name not in self._zones:
                     # update dbus zone
@@ -641,13 +816,8 @@ class FirewallConfig:
         os.remove("%s/%s.xml" % (ETC_FIREWALLD_ZONES, obj.name))
         del self._zones[obj.name]
 
-    def is_builtin_zone(self, obj):
-        if obj.default or obj.name in self._default_zones:
-            return True
-        return False
-
     def check_builtin_zone(self, obj):
-        if self.is_builtin_zone(obj):
+        if obj.builtin:
             raise FirewallError(BUILTIN_ZONE, "'%s' is built-in zone" % obj.name)
 
     def remove_zone(self, obj):
