@@ -75,7 +75,8 @@ class IcmpType(IO_Object):
         if item == "destination":
             for destination in config:
                 if destination not in [ "ipv4", "ipv6" ]:
-                    raise FirewallError(INVALID_DESTINATION, destination)
+                    raise FirewallError(INVALID_DESTINATION,
+                                 "'%s' not from {'ipv4'|'ipv6'}" % destination)
 
 # PARSER
 
@@ -103,12 +104,14 @@ class icmptype_ContentHandler(IO_Object_ContentHandler):
 def icmptype_reader(filename, path):
     icmptype = IcmpType()
     if not filename.endswith(".xml"):
-        raise FirewallError(INVALID_NAME, filename)
+        raise FirewallError(INVALID_NAME,
+                            "%s is missing .xml suffix" % filename)
     icmptype.name = filename[:-4]
     icmptype.check_name(icmptype.name)
     icmptype.filename = filename
     icmptype.path = path
-    icmptype.default = False if path.startswith(ETC_FIREWALLD) else True
+    icmptype.builtin = False if path.startswith(ETC_FIREWALLD) else True
+    icmptype.default = icmptype.builtin
     handler = icmptype_ContentHandler(icmptype)
     parser = sax.make_parser()
     parser.setContentHandler(handler)
@@ -135,6 +138,12 @@ def icmptype_writer(icmptype, path=None):
         except Exception as msg:
             raise IOError("Backup of '%s' failed: %s" % (name, msg))
 
+    dirpath = os.path.dirname(name)
+    if dirpath.startswith(ETC_FIREWALLD) and not os.path.exists(dirpath):
+        if not os.path.exists(ETC_FIREWALLD):
+            os.mkdir(ETC_FIREWALLD, 0o750)
+        os.mkdir(dirpath, 0o750)
+
     f = io.open(name, mode='wt', encoding='UTF-8')
     handler = IO_Object_XMLGenerator(f)
     handler.startDocument()
