@@ -26,7 +26,7 @@ import shutil
 
 from firewall.config import ETC_FIREWALLD
 from firewall.errors import *
-from firewall.functions import checkProtocol, check_address, \
+from firewall.functions import checkProtocol, \
                                checkIPnMask, checkIP6nMask, u2b_if_py2
 from firewall.core.io.io_object import *
 from firewall.core.logger import log
@@ -39,8 +39,9 @@ class Service(IO_Object):
         ( "ports", [ ( "", "" ), ], ),   # a(ss)
         ( "modules", [ "", ], ),         # as
         ( "destination", { "": "", }, ), # a{ss}
+        ( "protocols", [ "", ], ),       # as
         )
-    DBUS_SIGNATURE = '(sssa(ss)asa{ss})'
+    DBUS_SIGNATURE = '(sssa(ss)asa{ss}as)'
     ADDITIONAL_ALNUM_CHARS = [ "_", "-" ]
     PARSER_REQUIRED_ELEMENT_ATTRS = {
         "short": None,
@@ -50,6 +51,7 @@ class Service(IO_Object):
     PARSER_OPTIONAL_ELEMENT_ATTRS = {
         "service": [ "name", "version" ],
         "port": [ "port", "protocol" ],
+        "protocol": [ "value" ],
         "module": [ "name" ],
         "destination": [ "ipv4", "ipv6" ],
         }
@@ -60,6 +62,7 @@ class Service(IO_Object):
         self.short = ""
         self.description = ""
         self.ports = [ ]
+        self.protocols = [ ]
         self.modules = [ ]
         self.destination = { }
 
@@ -68,6 +71,7 @@ class Service(IO_Object):
         self.short = ""
         self.description = ""
         del self.ports[:]
+        del self.protocols[:]
         del self.modules[:]
         self.destination.clear()
 
@@ -81,27 +85,30 @@ class Service(IO_Object):
         self.ports = [(u2b_if_py2(po),u2b_if_py2(pr)) for (po,pr) in self.ports]
         self.modules = [u2b_if_py2(m) for m in self.modules]
         self.destination = {u2b_if_py2(k):u2b_if_py2(v) for k,v in self.destination.items()}
+        self.protocols = [u2b_if_py2(pr) for pr in self.protocols]
 
     def _check_config(self, config, item):
         if item == "ports":
             for port in config:
                 if port[0] != "":
                     check_port(port[0])
-                    check_protocol(port[1])
+                    check_tcpudp(port[1])
                 else:
                     # only protocol
-                    if not checkProtocol(port[1]):
-                        raise FirewallError(INVALID_PROTOCOL, port[1])
+                    check_protocol(port[1])
+
+        if item == "protocols":
+            for proto in config:
+                check_protocol(proto)
 
         elif item == "destination":
             for destination in config:
                 if destination not in [ "ipv4", "ipv6" ]:
                     raise FirewallError(INVALID_DESTINATION,
-                                    "'%s' not in {'ipv4'|'ipv6'}" % destination)
-                if not check_address(destination, config[destination]):
-                    raise FirewallError(INVALID_ADDR,
-                                        "'%s' is not valid %s address" % \
-                                        (config[destination], destination))
+                                        "'%s' not in {'ipv4'|'ipv6'}" % \
+                                        destination)
+                check_address(destination, config[destination])
+
         elif item == "modules":
             for module in config:
                 if not module.startswith("nf_conntrack_"):
@@ -117,7 +124,7 @@ class service_ContentHandler(IO_Object_ContentHandler):
         self.item.parser_check_element_attrs(name, attrs)
         if name == "service":
             if "name" in attrs:
-                log.warning("Ignoring deprecated attribute name='%s'" % 
+                log.warning("Ignoring deprecated attribute name='%s'",
                             attrs["name"])
             if "version" in attrs:
                 self.item.version = attrs["version"]
@@ -126,16 +133,49 @@ class service_ContentHandler(IO_Object_ContentHandler):
         elif name == "description":
             pass
         elif name == "port":
-            self.item.ports.append((attrs["port"], attrs["protocol"]))
+            if attrs["port"] != "":
+                check_port(attrs["port"])
+                check_tcpudp(attrs["protocol"])
+                entry = (attrs["port"], attrs["protocol"])
+                if entry not in self.item.ports:
+                    self.item.ports.append(entry)
+                else:
+                    log.warning("Port '%s/%s' already set, ignoring.",
+                                attrs["port"], attrs["protocol"])
+            else:
+                check_protocol(attrs["protocol"])
+                if attrs["protocol"] not in self.item.protocols:
+                    self.item.protocols.append(attrs["protocol"])
+                else:
+                    log.warning("Protocol '%s' already set, ignoring.",
+                                attrs["protocol"])
+        elif name == "protocol":
+            check_protocol(attrs["value"])
+            if attrs["value"] not in self.item.protocols:
+                self.item.protocols.append(attrs["value"])
+            else:
+                log.warning("Protocol '%s' already set, ignoring.",
+                            attrs["value"])
         elif name == "destination":
             for x in [ "ipv4", "ipv6" ]:
                 if x in attrs:
-                    if not check_address(x, attrs[x]):
-                        raise FirewallError(INVALID_ADDR,
-                                "'%s' is not valid %s address" % (attrs[x], x))
-                    self.item.destination[x] = attrs[x]
+                    check_address(x, attrs[x])
+                    if x in self.item.destination:
+                        log.warning("Destination address for '%s' already set, ignoring",
+                                    x)
+                    else:
+                        self.item.destination[x] = attrs[x]
         elif name == "module":
-            self.item.modules.append(attrs["name"])
+            if attrs["name"].startswith("nf_conntrack_") and \
+               len(attrs["name"].replace("nf_conntrack_", "")) > 0:
+                if attrs["name"] not in self.item.modules:
+                    self.item.modules.append(attrs["name"])
+                else:
+                    log.warning("Module '%s' already set, ignoring.",
+                                attrs["name"])
+            else:
+                log.warning("Invalid module '%s'", attrs["name"])
+
 
 def service_reader(filename, path):
     service = Service()
@@ -146,7 +186,8 @@ def service_reader(filename, path):
     service.check_name(service.name)
     service.filename = filename
     service.path = path
-    service.default = False if path.startswith(ETC_FIREWALLD) else True
+    service.builtin = False if path.startswith(ETC_FIREWALLD) else True
+    service.default = service.builtin
     handler = service_ContentHandler(service)
     parser = sax.make_parser()
     parser.setContentHandler(handler)
@@ -212,6 +253,12 @@ def service_writer(service, path=None):
         handler.simpleElement("port", { "port": port[0], "protocol": port[1] })
         handler.ignorableWhitespace("\n")
 
+    # protocols
+    for protocol in service.protocols:
+        handler.ignorableWhitespace("  ")
+        handler.simpleElement("protocol", { "value": protocol })
+        handler.ignorableWhitespace("\n")
+
     # modules
     for module in service.modules:
         handler.ignorableWhitespace("  ")
