@@ -25,7 +25,7 @@ import shutil
 
 from firewall.config import ETC_FIREWALLD
 from firewall.errors import *
-from firewall.functions import checkIP, uniqify, max_zone_name_len, u2b_if_py2
+from firewall.functions import checkIP, checkIP6, checkIPnMask, checkIP6nMask, checkInterface, uniqify, max_zone_name_len, u2b_if_py2, check_mac, portStr
 from firewall.core.base import DEFAULT_ZONE_TARGET, ZONE_TARGETS
 from firewall.core.io.io_object import *
 from firewall.core.rich import *
@@ -48,8 +48,9 @@ class Zone(IO_Object):
         ( "interfaces", [ "" ] ),                      # as
         ( "sources", [ "" ] ),                         # as
         ( "rules_str", [ "" ] ),                       # as
+        ( "protocols", [ "", ], ),                     # as
         )
-    DBUS_SIGNATURE = '(sssbsasa(ss)asba(ssss)asasas)'
+    DBUS_SIGNATURE = '(sssbsasa(ss)asba(ssss)asasasas)'
     ADDITIONAL_ALNUM_CHARS = [ "_", "-", "/" ]
     PARSER_REQUIRED_ELEMENT_ATTRS = {
         "short": None,
@@ -61,7 +62,7 @@ class Zone(IO_Object):
         "forward-port": [ "port", "protocol" ],
         "interface": [ "name" ],
         "rule": None,
-        "source": [ "address" ],
+        "source": None,
         "destination": [ "address" ],
         "protocol": [ "value" ],
         "log":  None,
@@ -69,6 +70,7 @@ class Zone(IO_Object):
         "accept": None,
         "reject": None,
         "drop": None,
+        "mark": [ "set" ],
         "limit": [ "value" ],
         }
     PARSER_OPTIONAL_ELEMENT_ATTRS = {
@@ -76,7 +78,7 @@ class Zone(IO_Object):
         "masquerade": [ "enabled" ],
         "forward-port": [ "to-port", "to-addr" ],
         "rule": [ "family" ],
-        "source": [ "invert", "family" ],
+        "source": [ "address", "mac", "invert", "family", "ipset" ],
         "destination": [ "invert" ],
         "log": [ "prefix", "level" ],
         "reject": [ "type" ],
@@ -98,6 +100,7 @@ class Zone(IO_Object):
         self.target = DEFAULT_ZONE_TARGET
         self.services = [ ]
         self.ports = [ ]
+        self.protocols = [ ]
         self.icmp_blocks = [ ]
         self.masquerade = False
         self.forward_ports = [ ]
@@ -116,6 +119,7 @@ class Zone(IO_Object):
         self.target = DEFAULT_ZONE_TARGET
         del self.services[:]
         del self.ports[:]
+        del self.protocols[:]
         del self.icmp_blocks[:]
         self.masquerade = False
         del self.forward_ports[:]
@@ -136,6 +140,7 @@ class Zone(IO_Object):
         self.target = u2b_if_py2(self.target)
         self.services = [u2b_if_py2(s) for s in self.services]
         self.ports = [(u2b_if_py2(po),u2b_if_py2(pr)) for (po,pr) in self.ports]
+        self.protocols = [u2b_if_py2(pr) for pr in self.protocols]
         self.icmp_blocks = [u2b_if_py2(i) for i in self.icmp_blocks]
         self.forward_ports = [(u2b_if_py2(p1),u2b_if_py2(p2),u2b_if_py2(p3),u2b_if_py2(p4)) for (p1,p2,p3,p4) in self.forward_ports]
         self.interfaces = [u2b_if_py2(i) for i in self.interfaces]
@@ -151,7 +156,7 @@ class Zone(IO_Object):
 
     def __setattr__(self, name, value):
         if name == "rules_str":
-            self.rules = [Rich_Rule(rule_str=str) for str in value]
+            self.rules = [Rich_Rule(rule_str=s) for s in value]
         else:
             object.__setattr__(self, name, value)
 
@@ -165,7 +170,10 @@ class Zone(IO_Object):
         elif item == "ports":
             for port in config:
                 check_port(port[0])
-                check_protocol(port[1])
+                check_tcpudp(port[1])
+        elif item == "protocols":
+            for proto in config:
+                check_protocol(proto)
         elif item == "icmp_blocks" and self.fw_config:
             existing_icmptypes = self.fw_config.get_icmptypes()
             for icmptype in config:
@@ -175,7 +183,7 @@ class Zone(IO_Object):
         elif item == "forward_ports":
             for fwd_port in config:
                 check_port(fwd_port[0])
-                check_protocol(fwd_port[1])
+                check_tcpudp(fwd_port[1])
                 if not fwd_port[2] and not fwd_port[3]:
                     raise FirewallError(INVALID_FORWARD,
                              "'%s' is missing to-port AND to-addr " % fwd_port)
@@ -188,6 +196,18 @@ class Zone(IO_Object):
         elif item == "target":
             if config not in ZONE_TARGETS:
                 raise FirewallError(INVALID_TARGET, config)
+        elif item == "interfaces":
+            for interface in config:
+                if not checkInterface(interface):
+                    raise FirewallError(INVALID_INTERFACE, interface)
+        elif item == "sources":
+            for source in config:
+                if not checkIPnMask(source) and not checkIP6nMask(source) and \
+                   not check_mac(source) and not source.startswith("ipset:"):
+                    raise FirewallError(INVALID_ADDR, source)
+        elif item == "rules_str":
+            for rule in config:
+                Rich_Rule(rule_str=rule)
 
     def check_name(self, name):
         super(Zone, self).check_name(name)
@@ -197,6 +217,9 @@ class Zone(IO_Object):
             raise FirewallError(INVALID_NAME, "'%s' can't end with '/'" % name)
         elif name.count('/') > 1:
             raise FirewallError(INVALID_NAME, "more than one '/' in '%s'" % name)
+        elif len(name) > max_zone_name_len():
+            raise FirewallError(INVALID_NAME,
+                                "'%s' has %d chars, max is %d" % (name, len(name), max_zone_name_len()))
 
     def combine(self, zone):
         self.combined = True
@@ -217,6 +240,9 @@ class Zone(IO_Object):
         for port in zone.ports:
             if port not in self.ports:
                 self.ports.append(port)
+        for proto in zone.protocols:
+            if proto not in self.protocols:
+                self.protocols.append(protocol)
         for icmp in zone.icmp_blocks:
             if icmp not in self.icmp_blocks:
                 self.icmp_blocks.append(icmp)
@@ -246,12 +272,12 @@ class zone_ContentHandler(IO_Object_ContentHandler):
 
         if name == "zone":
             if "name" in attrs:
-                log.warning("Ignoring deprecated attribute name='%s'" % 
+                log.warning("Ignoring deprecated attribute name='%s'",
                             attrs["name"])
             if "version" in attrs:
                 self.item.version = attrs["version"]
             if "immutable" in attrs:
-                log.warning("Ignoring deprecated attribute immutable='%s'" % 
+                log.warning("Ignoring deprecated attribute immutable='%s'",
                             attrs["immutable"])
             if "target" in attrs:
                 target = attrs["target"]
@@ -267,57 +293,84 @@ class zone_ContentHandler(IO_Object_ContentHandler):
         elif name == "service":
             if self._rule:
                 if self._rule.element:
-                    log.error('Invalid rule: More than one element, ignoring.')
+                    log.warning("Invalid rule: More than one element in rule '%s', ignoring.",
+                                str(self._rule))
                     self._rule_error = True
                     return
                 self._rule.element = Rich_Service(attrs["name"])
                 return
             if attrs["name"] not in self.item.services:
                 self.item.services.append(attrs["name"])
+            else:
+                log.warning("Service '%s' already set, ignoring.",
+                            attrs["name"])
+
         elif name == "port":
             if self._rule:
                 if self._rule.element:
-                    log.error('Invalid rule: More than one element, ignoring.')
+                    log.warning("Invalid rule: More than one element in rule '%s', ignoring.",
+                                str(self._rule))
                     self._rule_error = True
                     return
                 self._rule.element = Rich_Port(attrs["port"],
                                                attrs["protocol"])
                 return
-            # TODO: fix port string according to fw_zone.__port_id()
-            entry = (attrs["port"], attrs["protocol"])
+            check_port(attrs["port"])
+            check_tcpudp(attrs["protocol"])
+            entry = (portStr(attrs["port"], "-"), attrs["protocol"])
             if entry not in self.item.ports:
                 self.item.ports.append(entry)
+            else:
+                log.warning("Port '%s/%s' already set, ignoring.",
+                            attrs["port"], attrs["protocol"])
+
         elif name == "protocol":
             if self._rule:
                 if self._rule.element:
-                    log.error('Invalid rule: More than one element, ignoring.')
+                    log.warning("Invalid rule: More than one element in rule '%s', ignoring.",
+                                str(self._rule))
                     self._rule_error = True
                     return
                 self._rule.element = Rich_Protocol(attrs["value"])
             else:
-                log.error('Protocol allowed only in rule.')
+                check_protocol(attrs["value"])
+                if attrs["value"] not in self.item.protocols:
+                    self.item.protocols.append(attrs["value"])
+                else:
+                    log.warning("Protocol '%s' already set, ignoring.",
+                                attrs["value"])
         elif name == "icmp-block":
             if self._rule:
                 if self._rule.element:
-                    log.error('Invalid rule: More than one element, ignoring.')
+                    log.warning("Invalid rule: More than one element in rule '%s', ignoring.",
+                                str(self._rule))
                     self._rule_error = True
                     return
                 self._rule.element = Rich_IcmpBlock(attrs["name"])
                 return
             if attrs["name"] not in self.item.icmp_blocks:
                 self.item.icmp_blocks.append(attrs["name"])
+            else:
+                log.warning("icmp-block '%s' already set, ignoring.",
+                            attrs["name"])
+
         elif name == "masquerade":
             if "enabled" in attrs:
-                log.warning("Ignoring deprecated attribute enabled='%s'" %
+                log.warning("Ignoring deprecated attribute enabled='%s'",
                             attrs["enabled"])
             if self._rule:
                 if self._rule.element:
-                    log.error('Invalid rule: More than one element, ignoring.')
+                    log.warning("Invalid rule: More than one element in rule '%s', ignoring.",
+                                str(self._rule))
                     self._rule_error = True
                     return
                 self._rule.element = Rich_Masquerade()
             else:
-                self.item.masquerade = True
+                if self.item.masquerade:
+                    log.warning("Masquerade already set, ignoring.")
+                else:
+                    self.item.masquerade = True
+
         elif name == "forward-port":
             to_port = ""
             if "to-port" in attrs:
@@ -328,66 +381,112 @@ class zone_ContentHandler(IO_Object_ContentHandler):
 
             if self._rule:
                 if self._rule.element:
-                    log.error('Invalid rule: More than one element, ignoring.')
+                    log.warning("Invalid rule: More than one element in rule '%s', ignoring.",
+                                str(self._rule))
                     self._rule_error = True
                     return
                 self._rule.element = Rich_ForwardPort(attrs["port"],
                                                       attrs["protocol"],
                                                       to_port, to_addr)
                 return
-            # TODO: fix port string according to fw_zone.__forward_port_id()
-            entry = (attrs["port"], attrs["protocol"], to_port,
-                     to_addr)
+
+            check_port(attrs["port"])
+            check_tcpudp(attrs["protocol"])
+            if to_port:
+                check_port(to_port)
+            if to_addr:
+                if not checkIP(to_addr):
+                    raise FirewallError(INVALID_ADDR,
+                                        "to-addr '%s' is not a valid address" \
+                                        % to_addr)
+            entry = (portStr(attrs["port"], "-"), attrs["protocol"],
+                     portStr(to_port, "-"), str(to_addr))
             if entry not in self.item.forward_ports:
                 self.item.forward_ports.append(entry)
+            else:
+                log.warning("Forward port %s/%s%s%s already set, ignoring.",
+                            attrs["port"], attrs["protocol"],
+                            " >%s" % to_port if to_port else "",
+                            " @%s" % to_addr if to_addr else "")
 
         elif name == "interface":
             if self._rule:
-                log.error('Invalid rule: interface use in rule.')
+                log.warning('Invalid rule: interface use in rule.')
                 self._rule_error = True
                 return
             # zone bound to interface
-            if not "name" in attrs:
-                log.error('Invalid interface: Name missing.')
+            if "name" not in attrs:
+                log.warning('Invalid interface: Name missing.')
                 self._rule_error = True
                 return
-            name = attrs["name"]
-            if name not in self.item.interfaces:
-                self.item.interfaces.append(name)
-            
+            attrs["name"]
+            if attrs["name"] not in self.item.interfaces:
+                self.item.interfaces.append(attrs["name"])
+            else:
+                log.warning("Interface '%s' already set, ignoring.",
+                            attrs["name"])
+
         elif name == "source":
             if self._rule:
                 if self._rule.source:
-                    log.error('Invalid rule: More than one source')
+                    log.warning("Invalid rule: More than one source in rule '%s', ignoring.",
+                                str(self._rule))
                     self._rule_error = True
                     return
                 invert = False
                 if "invert" in attrs and \
                         attrs["invert"].lower() in [ "yes", "true" ]:
                     invert = True
-                self._rule.source = Rich_Source(attrs["address"], invert)
+                addr = mac = ipset = None
+                if "address" in attrs:
+                    addr = attrs["address"]
+                if "mac" in attrs:
+                    mac = attrs["mac"]
+                if "ipset" in attrs:
+                    ipset = attrs["ipset"]
+                self._rule.source = Rich_Source(addr, mac, ipset, invert=invert)
                 return
             # zone bound to source
-            if not "address" in attrs:
-                log.error('Invalid source: Address missing.')
+            if "address" not in attrs and not "ipset" in attrs:
+                log.warning('Invalid source: No address no ipset.')
+                return
+            if "address" in attrs and "ipset" in attrs:
+                log.warning('Invalid source: Address and ipset.')
                 return
             if "family" in attrs:
-                log.warning("Ignoring deprecated attribute family='%s'" %
+                log.warning("Ignoring deprecated attribute family='%s'",
                             attrs["family"])
             if "invert" in attrs:
-                log.error('Invalid source: Invertion not allowed here.')
+                log.warning('Invalid source: Invertion not allowed here.')
                 return
-            entry = attrs["address"]
-            if entry not in self.item.sources:
-                self.item.sources.append(entry)
+            if "address" in attrs:
+                if not checkIPnMask(attrs["address"]) and \
+                   not checkIP6nMask(attrs["address"]) and \
+                   not check_mac(attrs["address"]):
+                    raise FirewallError(INVALID_ADDR, source)
+            if "ipset" in attrs:
+                entry = "ipset:%s" % attrs["ipset"]
+                if entry not in self.item.sources:
+                    self.item.sources.append(entry)
+                else:
+                    log.warning("Source '%s' already set, ignoring.",
+                                attrs["address"])
+            if "address" in attrs:
+                entry = attrs["address"]
+                if entry not in self.item.sources:
+                    self.item.sources.append(entry)
+                else:
+                    log.warning("Source '%s' already set, ignoring.",
+                                attrs["address"])
 
         elif name == "destination":
             if not self._rule:
-                log.error('Invalid rule: Destination outside of rule')
+                log.warning('Invalid rule: Destination outside of rule')
                 self._rule_error = True
                 return
             if self._rule.destination:
-                log.error('Invalid rule: More than one destination')
+                log.warning("Invalid rule: More than one destination in rule '%s', ignoring.",
+                            str(self._rule))
                 return
             invert = False
             if "invert" in attrs and \
@@ -396,39 +495,42 @@ class zone_ContentHandler(IO_Object_ContentHandler):
             self._rule.destination = Rich_Destination(attrs["address"],
                                                       invert)
 
-        elif name in [ "accept", "reject", "drop" ]:
+        elif name in [ "accept", "reject", "drop", "mark" ]:
             if not self._rule:
-                log.error('Invalid rule: Action outside of rule')
+                log.warning('Invalid rule: Action outside of rule')
                 self._rule_error = True
                 return
             if self._rule.action:
-                log.error('Invalid rule: More than one action')
+                log.warning('Invalid rule: More than one action')
                 self._rule_error = True
                 return
             if name == "accept":
                 self._rule.action = Rich_Accept()
-            if name == "reject":
+            elif name == "reject":
                 _type = None
                 if "type" in attrs:
                     _type = attrs["type"]
                 self._rule.action = Rich_Reject(_type)
-            if name == "drop":
+            elif name == "drop":
                 self._rule.action = Rich_Drop()
+            elif name == "mark":
+                _set = attrs["set"]
+                self._rule.action = Rich_Mark(_set)
             self._limit_ok = self._rule.action
 
         elif name == "log":
             if not self._rule:
-                log.error('Invalid rule: Log outside of rule')
+                log.warning('Invalid rule: Log outside of rule')
                 return
             if self._rule.log:
-                log.error('Invalid rule: More than one log')
+                log.warning('Invalid rule: More than one log')
                 return
             level = None
             if "level" in attrs:
                 level = attrs["level"]
                 if level not in [ "emerg", "alert", "crit", "error",
                                   "warning", "notice", "info", "debug" ]:
-                    log.error('Invalid rule: Invalid log level')
+                    log.warning('Invalid rule: Invalid log level')
                     self._rule_error = True
                     return
             prefix = attrs["prefix"] if "prefix" in attrs else None
@@ -437,10 +539,11 @@ class zone_ContentHandler(IO_Object_ContentHandler):
 
         elif name == "audit":
             if not self._rule:
-                log.error('Invalid rule: Audit outside of rule')
+                log.warning('Invalid rule: Audit outside of rule')
                 return
             if self._rule.audit:
-                log.error('Invalid rule: More than one audit')
+                log.warning("Invalid rule: More than one audit in rule '%s', ignoring.",
+                            str(self._rule))
                 self._rule_error = True
                 return            
             self._rule.audit = Rich_Audit()
@@ -451,27 +554,27 @@ class zone_ContentHandler(IO_Object_ContentHandler):
             if "family" in attrs:
                 family = attrs["family"]
                 if family not in [ "ipv4", "ipv6" ]:
-                    log.error('Invalid rule: Rule family "%s" invalid' % 
-                              attrs["family"])
+                    log.warning('Invalid rule: Rule family "%s" invalid',
+                                attrs["family"])
                     self._rule_error = True
                     return
             self._rule = Rich_Rule(family)
-            self.item.rules.append(self._rule)
 
         elif name == "limit":
             if not self._limit_ok:
-                log.error('Invalid rule: Limit outside of action, log and audit')
+                log.warning('Invalid rule: Limit outside of action, log and audit')
                 self._rule_error = True
                 return
             if self._limit_ok.limit:
-                log.error('Invalid rule: More than one limit')
+                log.warning("Invalid rule: More than one limit in rule '%s', ignoring.",
+                            str(self._rule))
                 self._rule_error = True
                 return
             value = attrs["value"]
             self._limit_ok.limit = Rich_Limit(value)
 
         else:
-            log.error('Unknown XML element %s' % name)
+            log.warning("Unknown XML element '%s'", name)
             return
 
     def endElement(self, name):
@@ -482,13 +585,17 @@ class zone_ContentHandler(IO_Object_ContentHandler):
                 try:
                     self._rule.check()
                 except Exception as e:
-                    log.error("%s: %s" % (e, str(self._rule)))
-                    self._rule_error = True
-            if self._rule_error and self._rule in self.item.rules:
-                self.item.rules.remove(self._rule)
+                    log.warning("%s: %s", e, str(self._rule))
+                else:
+                    if str(self._rule) not in \
+                       [ str(x) for x in self.item.rules ]:
+                        self.item.rules.append(self._rule)
+                    else:
+                        log.warning("Rule '%s' already set, ignoring.",
+                                    str(self._rule))
             self._rule = None
             self._rule_error = False
-        elif name in [ "accept", "reject", "drop", "log", "audit" ]:
+        elif name in [ "accept", "reject", "drop", "mark", "log", "audit" ]:
             self._limit_ok = None
 
 def zone_reader(filename, path):
@@ -503,7 +610,8 @@ def zone_reader(filename, path):
     zone.check_name(zone.name)
     zone.filename = filename
     zone.path = path
-    zone.default = False if path.startswith(ETC_FIREWALLD) else True
+    zone.builtin = False if path.startswith(ETC_FIREWALLD) else True
+    zone.default = zone.builtin
     handler = zone_ContentHandler(zone)
     parser = sax.make_parser()
     parser.setContentHandler(handler)
@@ -574,7 +682,10 @@ def zone_writer(zone, path=None):
     # source
     for source in uniqify(zone.sources):
         handler.ignorableWhitespace("  ")
-        handler.simpleElement("source", { "address": source })
+        if "ipset:" in source:
+            handler.simpleElement("source", { "ipset": source[6:] })
+        else:
+            handler.simpleElement("source", { "address": source })
         handler.ignorableWhitespace("\n")
 
     # services
@@ -589,6 +700,12 @@ def zone_writer(zone, path=None):
         handler.simpleElement("port", { "port": port[0], "protocol": port[1] })
         handler.ignorableWhitespace("\n")
 
+    # protocols
+    for protocol in uniqify(zone.protocols):
+        handler.ignorableWhitespace("  ")
+        handler.simpleElement("protocol", { "value": protocol })
+        handler.ignorableWhitespace("\n")
+
     # icmp-blocks
     for icmp in uniqify(zone.icmp_blocks):
         handler.ignorableWhitespace("  ")
@@ -623,7 +740,13 @@ def zone_writer(zone, path=None):
 
         # source
         if rule.source:
-            attrs = { "address": rule.source.addr }
+            attrs = { }
+            if rule.source.addr:
+                attrs["address"] = rule.source.addr
+            if rule.source.mac:
+                attrs["mac"] = rule.source.mac
+            if rule.source.ipset:
+                attrs["ipset"] = rule.source.ipset
             if rule.source.invert:
                 attrs["invert"] = "True"
             handler.ignorableWhitespace("    ")
@@ -668,7 +791,7 @@ def zone_writer(zone, path=None):
                 if rule.element.to_address != "":
                     attrs["to-addr"] = rule.element.to_address
             else:
-                log.error('Unknown element "%s"' % type(rule.element))
+                log.warning("Unknown element '%s'", type(rule.element))
 
             handler.ignorableWhitespace("    ")
             handler.simpleElement(element, attrs)
@@ -724,8 +847,11 @@ def zone_writer(zone, path=None):
                     attrs["type"] = rule.action.type
             elif type(rule.action) == Rich_Drop:
                 action = "drop"
+            elif type(rule.action) == Rich_Mark:
+                action = "mark"
+                attrs["set"] = rule.action.set
             else:
-                log.error('Unknown action "%s"' % type(rule.action))
+                log.warning("Unknown action '%s'", type(rule.action))
             if rule.action.limit:
                 handler.ignorableWhitespace("    ")
                 handler.startElement(action, attrs)
