@@ -23,11 +23,8 @@ import os.path
 
 from firewall.core.prog import runProg
 from firewall.core.logger import log
-
-COMMAND = {
-    "ipv4": "/sbin/iptables",
-    "ipv6": "/sbin/ip6tables",
-}
+from firewall.functions import tempFile, readfile
+from firewall.config import COMMANDS
 
 PROC_IPxTABLE_NAMES = {
     "ipv4": "/proc/net/ip_tables_names",
@@ -53,6 +50,7 @@ ICMP = {
 }
 
 DEFAULT_RULES = { }
+LOG_RULES = { }
 OUR_CHAINS = {} # chains created by firewalld
 
 DEFAULT_RULES["security"] = [ ]
@@ -132,22 +130,31 @@ DEFAULT_RULES["filter"] = [
 
     "-I OUTPUT 1 -j OUTPUT_direct",
 ]
+
+LOG_RULES["filter"] = [
+    "-I INPUT 7 -m conntrack --ctstate INVALID %%LOGTYPE%% -j LOG --log-prefix 'STATE_INVALID_DROP: '",
+    "-I INPUT 9 %%LOGTYPE%% -j LOG --log-prefix 'FINAL_REJECT: '",
+
+    "-I FORWARD 9 -m conntrack --ctstate INVALID %%LOGTYPE%% -j LOG --log-prefix 'STATE_INVALID_DROP: '",
+    "-I FORWARD 11 %%LOGTYPE%% -j LOG --log-prefix 'FINAL_REJECT: '",
+]
+
 OUR_CHAINS["filter"] = set(["INPUT_direct", "INPUT_ZONES_SOURCE", "INPUT_ZONES",
-                          "FORWARD_direct", "FORWARD_IN_ZONES_SOURCE",
-                          "FORWARD_IN_ZONES", "FORWARD_OUT_ZONES_SOURCE",
-                          "FORWARD_OUT_ZONES", "OUTPUT_direct"])
+                            "FORWARD_direct", "FORWARD_IN_ZONES_SOURCE",
+                            "FORWARD_IN_ZONES", "FORWARD_OUT_ZONES_SOURCE",
+                            "FORWARD_OUT_ZONES", "OUTPUT_direct"])
 
-
-class ip4tables:
+class ip4tables(object):
     ipv = "ipv4"
 
     def __init__(self):
-        self._command = COMMAND[self.ipv]
+        self._command = COMMANDS[self.ipv]
+        self._restore_command = COMMANDS["%s-restore" % self.ipv]
         self.wait_option = self._detect_wait_option()
 
     def __run(self, args):
         # convert to string list
-        if self.wait_option:
+        if self.wait_option and self.wait_option not in args:
             _args = [self.wait_option] + ["%s" % item for item in args]
         else:
             _args = ["%s" % item for item in args]
@@ -158,6 +165,61 @@ class ip4tables:
                                                      " ".join(_args), ret))
         return ret
 
+    def set_rules(self, rules, flush=False):
+        temp_file = tempFile()
+
+        table = "filter"
+        table_rules = { }
+        for rule in rules:
+            try:
+                i = rule.index("-t")
+            except:
+                pass
+            else:
+                if len(rule) >= i+1:
+                    rule.pop(i)
+                    table = rule.pop(i)
+
+            table_rules.setdefault(table, []).append(rule)
+
+        for table in table_rules:
+            temp_file.write("*%s\n" % table)
+            for rule in table_rules[table]:
+                temp_file.write(" ".join(rule) + "\n")
+            temp_file.write("COMMIT\n")
+
+        temp_file.close()
+
+        stat = os.stat(temp_file.name)
+        log.debug2("%s: %s %s", self.__class__, self._restore_command,
+                   "%s: %d" % (temp_file.name, stat.st_size))
+        args = [ ]
+        if not flush:
+            args.append("-n")
+
+        (status, ret) = runProg(self._restore_command, args,
+                                stdin=temp_file.name)
+
+        if log.getDebugLogLevel() > 2:
+            try:
+                lines = readfile(temp_file.name)
+            except:
+                pass
+            else:
+                i = 1
+                for line in readfile(temp_file.name):
+                    log.debug3("%8d: %s" % (i, line), nofmt=1, nl=0)
+                    if not line.endswith("\n"):
+                        log.debug3("", nofmt=1)
+                    i += 1
+
+        os.unlink(temp_file.name)
+
+        if status != 0:
+            raise ValueError("'%s %s' failed: %s" % (self._restore_command,
+                                                     " ".join(args), ret))
+        return ret
+
     def set_rule(self, rule):
         return self.__run(rule)
 
@@ -172,7 +234,7 @@ class ip4tables:
         tables = [ table ] if table else BUILT_IN_CHAINS.keys()
         for table in tables:
             try:
-                self.__run(["-t", table, "-L"])
+                self.__run(["-t", table, "-L", "-n"])
                 ret.append(table)
             except ValueError:
                 log.debug1("%s table '%s' does not exist (or not enough permission to check)." % (self.ipv, table))
@@ -194,26 +256,32 @@ class ip4tables:
 
     def _detect_wait_option(self):
         wait_option = ""
-        (status, ret) = runProg(self._command, ["-w", "-L"])  # since iptables-1.4.20
+        (status, ret) = runProg(self._command, ["-w", "-L", "-n"])  # since iptables-1.4.20
         if status == 0:
             wait_option = "-w"  # wait for xtables lock
-            (status, ret) = runProg(self._command, ["-w2", "-L"])  # since iptables > 1.4.21
+            (status, ret) = runProg(self._command, ["-w2", "-L", "-n"])  # since iptables > 1.4.21
             if status == 0:
                 wait_option = "-w2"  # wait max 2 seconds
             log.debug2("%s: %s will be using %s option.", self.__class__, self._command, wait_option)
 
         return wait_option
 
-    def flush(self):
+    def flush(self, individual=False):
         tables = self.used_tables()
+        rules = [ ]
         for table in tables:
             # Flush firewall rules: -F
             # Delete firewall chains: -X
             # Set counter to zero: -Z
             for flag in [ "-F", "-X", "-Z" ]:
-                self.__run([ "-t", table, flag ])
+                if individual:
+                    self.__run([ "-t", table, flag ])
+                else:
+                    rules.append([ "-t", table, flag ])
+        if len(rules) > 0:
+            self.set_rules(rules)
 
-    def set_policy(self, policy, which="used"):
+    def set_policy(self, policy, which="used", individual=False):
         if which == "used":
             tables = self.used_tables()
         else:
@@ -222,66 +290,18 @@ class ip4tables:
         if "nat" in tables:
             tables.remove("nat") # nat can not set policies in nat table
 
+        rules = [ ]
         for table in tables:
             for chain in BUILT_IN_CHAINS[table]:
-                self.__run([ "-t", table, "-P", chain, policy ])
+                if individual:
+                    self.__run([ "-t", table, "-P", chain, policy ])
+                else:
+                    rules.append([ "-t", table, "-P", chain, policy ])
+        if len(rules) > 0:
+            self.set_rules(rules)
 
 class ip6tables(ip4tables):
     ipv = "ipv6"
 
 ip4tables_available_tables = ip4tables().available_tables()
 ip6tables_available_tables = ip6tables().available_tables()
-
-#class ipXtables:
-#    def __init__(self, ipv4=True, ipv6=True):
-#        self.ip4tables = self.ip6tables = None
-#        if ipv4:
-#            self.ip4tables = ip4tables()
-#        if ipv6:
-#            self.ip6tables = ip6tables()
-#        if not ipv4 and not ipv6:
-#            raise ValueError("ipv4 and ipv6 disabled.")
-
-#    def append_rule(self, rule):
-#        if self.ip4tables:
-#            self.ip4tables.append_rule(rule)
-#        try:
-#            if self.ip6tables:
-#                self.ip6tables.append_rule(rule)
-#        except Exception:
-#            if self.ip4tables:
-#                self.ip4tables.delete_rule(rule)
-#            raise
-
-#    def delete_rule(self, rule):
-#        if self.ip4tables:
-#            self.ip4tables.delete_rule(rule)
-#        try:
-#            if self.ip6tables:
-#                self.ip6tables.delete_rule(rule)
-#        except Exception:
-#            if self.ip4tables:
-#                self.ip4tables.append_rule(rule)
-#            raise
-
-#    def used_tables(self):
-#        tables = [ ]
-#        if ip4tables:
-#            tables += self.ip4tables.used_tables()
-#        if ip6tables:
-#            tables += self.ip6tables.used_tables()
-#        return tables
-
-#    def flush(self):
-#        if ip4tables:
-#            self.ip4tables.flush()
-#        if ip6tables:
-#            self.ip6tables.flush()
-        # TODO: in case off error state is inconsistent
-
-#    def set_policy(self, policy, which="used"):
-#        if ip4tables:
-#            self.ip4tables.set_policy(policy, which)
-#        if ip6tables:
-#            self.ip6tables.set_policy(policy, which)
-        # TODO: in case off error state is inconsistent
