@@ -21,14 +21,17 @@
 
 import os
 
-def runProg(prog, argv=[ ]):
+def runProg(prog, argv=[ ], stdin=None):
     args = [ prog ] + argv
 
     (rfd, wfd) = os.pipe()
     pid = os.fork()
     if pid == 0:
         try:
-            fd = os.open("/dev/null", os.O_RDONLY)
+            if stdin != None:
+                fd = os.open(stdin, os.O_RDONLY)
+            else:
+                fd = os.open("/dev/null", os.O_RDONLY)
             if fd != 0:
                 os.dup2(fd, 0)
                 os.close(fd)
@@ -42,7 +45,7 @@ def runProg(prog, argv=[ ]):
             os._exit(255)
     os.close(wfd)
 
-    cret = ""
+    cret = b''
     cout = os.read(rfd, 8192)
     while cout:
         cret += cout
@@ -50,4 +53,5 @@ def runProg(prog, argv=[ ]):
     os.close(rfd)
     (cpid, status) = os.waitpid(pid, 0)
 
-    return (status, cret.rstrip())
+    cret = cret.rstrip().decode('utf-8', 'replace')
+    return (status, cret)
