@@ -21,6 +21,7 @@
 
 from firewall import functions
 from firewall.errors import *
+from firewall.core.ipset import check_ipset_name
 
 REJECT_TYPES = {
     "ipv4": ["icmp-host-prohibited", "host-prohib", "icmp-net-unreachable", "net-unreach", "icmp-host-unreachable", "host-unreach", "icmp-port-unreachable", "port-unreach", "icmp-proto-unreachable", "proto-unreach", "icmp-net-prohibited", "net-prohib", "tcp-reset", "tcp-rst", "icmp-admin-prohibited", "admin-prohib"],
@@ -28,15 +29,34 @@ REJECT_TYPES = {
 }
 
 class Rich_Source(object):
+    def __init__(self, addr, mac, ipset, invert=False):
+        self.addr = addr
+        if self.addr == "":
+            self.addr = None
+        self.mac = mac
+        if self.mac == "" or self.mac == None:
+            self.mac = None
+        elif self.mac != None:
+            self.mac = self.mac.upper()
+        self.ipset = ipset
+        if self.ipset == "":
+            self.ipset = None
+        self.invert = invert
+
+    def __str__(self):
+        if self.addr:
+            x = ' address="%s"' % self.addr
+        elif self.mac:
+            x = ' mac="%s"' % self.mac
+        elif self.ipset:
+            x = ' ipset="%s"' % self.ipset
+        return 'source%s%s' % (" NOT" if self.invert else "", x)
+
+class Rich_Destination(object):
     def __init__(self, addr, invert=False):
         self.addr = addr
         self.invert = invert
 
-    def __str__(self):
-        return 'source %saddress="%s"' % ("not " if self.invert else "",
-                                          self.addr)
-
-class Rich_Destination(Rich_Source):
     def __str__(self):
         return 'destination %saddress="%s"' % ("not " if self.invert else "",
                                                self.addr)
@@ -146,6 +166,34 @@ class Rich_Drop(Rich_Accept):
         return "drop%s" % (" %s" % self.limit if self.limit else "")
 
 
+class Rich_Mark(object):
+    def __init__(self, _set, limit=None):
+        self.set = _set
+        self.limit = limit
+
+    def __str__(self):
+        return "mark set=%s%s" % (self.set,
+                                  " %s" % self.limit if self.limit else "")
+
+    def check(self):
+        if self.set is not None:
+            x = self.set
+        else:
+            raise FirewallError(INVALID_MARK, "no value set")
+
+        if "/" in x:
+            splits = x.split("/")
+            if len(splits) != 2:
+                raise FirewallError(INVALID_MARK, x)
+            if not functions.checkUINT32(splits[0]) or \
+               not functions.checkUINT32(splits[1]):
+                # value and mask are uint32
+                raise FirewallError(INVALID_MARK, x)
+        else:
+            if not functions.checkUINT32(x):
+                # value is uint32
+                raise FirewallError(INVALID_MARK, x)
+
 class Rich_Limit(object):
     def __init__(self, value):
         self.value = value
@@ -167,7 +215,7 @@ class Rich_Limit(object):
         except:
             raise FirewallError(INVALID_LIMIT, self.value)
 
-        if rate < 0 or duration not in [ "s", "m", "h", "d" ]:
+        if rate < 1 or duration not in [ "s", "m", "h", "d" ]:
             raise FirewallError(INVALID_LIMIT, self.value)
 
         mult = 1
@@ -183,6 +231,10 @@ class Rich_Limit(object):
         if 10000 * mult / rate == 0:
             raise FirewallError(INVALID_LIMIT, "%s too fast" % self.value)
 
+        if rate == 1 and duration == "d":
+            # iptables (v1.4.21) doesn't accept 1/d
+            raise FirewallError(INVALID_LIMIT, "%s too slow" % self.value)
+
     def __str__(self):
         return 'limit value="%s"' % (self.value)
 
@@ -248,15 +300,17 @@ class Rich_Rule(object):
             #print ("in_elements: ", in_elements)
             #print ("index: %s, element: %s, attribute: %s=%s" % (index, element, attr_name, attr_value))
             if attr_name:     # attribute
-                if attr_name not in ['family', 'address', 'invert', 'value',
+                if attr_name not in ['family', 'address', 'mac', 'ipset',
+                                     'invert', 'value',
                                      'port', 'protocol', 'to-port', 'to-addr',
-                                     'name', 'prefix', 'level', 'type']:
+                                     'name', 'prefix', 'level', 'type',
+                                     'set']:
                     raise FirewallError(INVALID_RULE, "bad attribute '%s'" % attr_name)
             else:             # element
                 if element in ['rule', 'source', 'destination', 'protocol',
                                'service', 'port', 'icmp-block', 'masquerade',
                                'forward-port', 'log', 'audit',
-                               'accept', 'drop', 'reject', 'limit', 'not', 'NOT', 'EOL']:
+                               'accept', 'drop', 'reject', 'mark', 'limit', 'not', 'NOT', 'EOL']:
                     if element == 'source' and self.source:
                         raise FirewallError(INVALID_RULE, "more than one 'source' element")
                     elif element == 'destination' and self.destination:
@@ -268,7 +322,7 @@ class Rich_Rule(object):
                         raise FirewallError(INVALID_RULE, "more than one 'log' element")
                     elif element == 'audit' and self.audit:
                         raise FirewallError(INVALID_RULE, "more than one 'audit' element")
-                    elif element in ['accept', 'drop', 'reject'] and self.action:
+                    elif element in ['accept', 'drop', 'reject', 'mark'] and self.action:
                         raise FirewallError(INVALID_RULE, "more than one 'action' element. There cannot be both '%s' and '%s' in one rule." % (element, self.action))
                 else:
                     raise FirewallError(INVALID_RULE, "unknown element %s" % element)
@@ -299,12 +353,12 @@ class Rich_Rule(object):
                 else:
                     in_elements.append(element) # push into stack
             elif in_element == 'source':
-                if attr_name in ['address', 'invert']:
+                if attr_name in ['address', 'mac', 'ipset', 'invert']:
                     attrs[attr_name] = attr_value
                 elif element in ['not', 'NOT']:
                     attrs['invert'] = True
                 else:
-                    self.source = Rich_Source(attrs.get('address'), attrs.get('invert'))
+                    self.source = Rich_Source(attrs.get('address'), attrs.get('mac'), attrs.get('ipset'), attrs.get('invert'))
                     in_elements.pop() # source
                     attrs.clear()
                     index = index -1 # return token to input
@@ -401,6 +455,17 @@ class Rich_Rule(object):
                     in_elements.pop() # accept
                     attrs.clear()
                     index = index -1 # return token to input
+            elif in_element == 'mark':
+                if attr_name == 'set':
+                    attrs[attr_name] = attr_value
+                elif element == 'limit':
+                    in_elements.append('limit')
+                else:
+                    self.action = Rich_Mark(attrs.get('set'),
+                                            attrs.get('limit'))
+                    in_elements.pop() # accept
+                    attrs.clear()
+                    index = index -1 # return token to input
             elif in_element == 'limit':
                 if attr_name == 'value':
                     attrs['limit'] = Rich_Limit(attr_value)
@@ -416,7 +481,8 @@ class Rich_Rule(object):
         if self.family is not None and self.family not in [ "ipv4", "ipv6" ]:
             raise FirewallError(INVALID_FAMILY, self.family)
         if self.family is None:
-            if self.source is not None or self.destination is not None:
+            if (self.source is not None and self.source.addr is not None) or \
+               self.destination is not None:
                 raise FirewallError(MISSING_FAMILY)
             if type(self.element) == Rich_ForwardPort:
                 raise FirewallError(MISSING_FAMILY)
@@ -438,12 +504,24 @@ class Rich_Rule(object):
 
         # source
         if self.source is not None:
-            if self.family is None:
-                raise FirewallError(INVALID_FAMILY)
-            if self.source.addr is None or \
-                    not functions.check_address(self.family,
-                                                self.source.addr):
-                raise FirewallError(INVALID_ADDR, str(self.source.addr))
+            if self.source.addr is not None:
+                if self.family is None:
+                    raise FirewallError(INVALID_FAMILY)
+                if self.source.mac is not None:
+                    raise FirewallError(INVALID_RULE, "address and mac")
+                if not functions.check_address(self.family, self.source.addr):
+                    raise FirewallError(INVALID_ADDR, str(self.source.addr))
+
+            elif self.source.mac is not None:
+                if not functions.check_mac(self.source.mac):
+                    raise FirewallError(INVALID_MAC, str(self.source.mac))
+
+            elif self.source.ipset is not None:
+                if not check_ipset_name(self.source.ipset):
+                    raise FirewallError(INVALID_IPSET, str(self.source.ipset))
+
+            else:
+                raise FirewallError(INVALID_RULE, "invalid source")
 
         # destination
         if self.destination is not None:
@@ -465,7 +543,7 @@ class Rich_Rule(object):
         elif type(self.element) == Rich_Port:
             if not functions.check_port(self.element.port):
                 raise FirewallError(INVALID_PORT, self.element.port)
-            if not self.element.protocol in [ "tcp", "udp" ]:
+            if self.element.protocol not in [ "tcp", "udp" ]:
                 raise FirewallError(INVALID_PROTOCOL, self.element.protocol)
 
         # protocol
@@ -475,11 +553,10 @@ class Rich_Rule(object):
 
         # masquerade
         elif type(self.element) == Rich_Masquerade:
-            if self.destination is not None:
-                raise FirewallError(INVALID_RULE, "masquerade and destination")
             if self.action is not None:
                 raise FirewallError(INVALID_RULE, "masquerade and action")
-
+            if self.source is not None and self.source.mac is not None:
+                raise FirewallError(INVALID_RULE, "masquerade and mac source")
 
         # icmp-block
         elif type(self.element) == Rich_IcmpBlock:
@@ -494,7 +571,7 @@ class Rich_Rule(object):
         elif type(self.element) == Rich_ForwardPort:
             if not functions.check_port(self.element.port):
                 raise FirewallError(INVALID_PORT, self.element.port)
-            if not self.element.protocol in [ "tcp", "udp" ]:
+            if self.element.protocol not in [ "tcp", "udp" ]:
                 raise FirewallError(INVALID_PROTOCOL, self.element.protocol)
             if self.element.to_port == "" and self.element.to_address == "":
                 raise FirewallError(INVALID_PORT, self.element.to_port)
@@ -537,6 +614,8 @@ class Rich_Rule(object):
         if self.action is not None:
             if type(self.action) == Rich_Reject:
                 self.action.check(self.family)
+            elif type(self.action) == Rich_Mark:
+                self.action.check()
 
             if self.action.limit is not None:
                 self.action.limit.check()
