@@ -24,7 +24,9 @@ import os.path
 import shlex, pipes
 import string
 import sys
+import tempfile
 from firewall.core.logger import log
+from firewall.config import FIREWALLD_TEMPDIR
 
 PY2 = sys.version < '3'
 
@@ -42,7 +44,7 @@ def getPortID(port):
             port = port.strip()
         try:
             id = int(port)
-        except:
+        except ValueError:
             try:
                 id = socket.getservbyname(port)
             except:
@@ -131,7 +133,7 @@ def checkIP(ip):
 
     try:
         socket.inet_pton(socket.AF_INET, ip)
-    except socket.error as err:
+    except socket.error:
         return False
     return True
 
@@ -144,7 +146,7 @@ def checkIP6(ip):
 
     try:
         socket.inet_pton(socket.AF_INET6, ip)
-    except socket.error as err:
+    except socket.error:
         return False
     return True
 
@@ -163,7 +165,7 @@ def checkIPnMask(ip):
         else:
             try:
                 i = int(mask)
-            except:
+            except ValueError:
                 return False
             if i < 0 or i > 32:
                 return False
@@ -181,7 +183,7 @@ def checkIP6nMask(ip):
     if mask:
         try:
             i = int(mask)
-        except:
+        except ValueError:
             return False
         if i < 0 or i > 128:
             return False
@@ -191,7 +193,7 @@ def checkIP6nMask(ip):
 def checkProtocol(protocol):
     try:
         i = int(protocol)
-    except:
+    except ValueError:
         # string
         try:
             socket.getprotobyname(protocol)
@@ -222,6 +224,16 @@ def checkInterface(iface):
     #    return False
     return True
 
+def checkUINT32(val):
+    try:
+        x = int(val, 0)
+    except:
+        return False
+    else:
+        if x >= 0 and x <= 4294967295:
+            return True
+    return False
+
 def firewalld_is_active():
     """ Check if firewalld is active
 
@@ -251,14 +263,23 @@ def firewalld_is_active():
 
     return False
 
+def tempFile():
+    try:
+        return tempfile.NamedTemporaryFile(mode='wt', prefix="temp.",
+                                           dir=FIREWALLD_TEMPDIR, delete=False)
+    except Exception as msg:
+        log.error("Failed to create temporary file: %s" % msg)
+        raise
+    return None
+
 def readfile(filename):
+    i = 1
     try:
         with open(filename, "r") as f:
-            line = "".join(f.readlines())
+            return f.readlines()
     except Exception as e:
         log.error('Failed to read file "%s": %s' % (filename, e))
-        return None
-    return line
+    return None
 
 def writefile(filename, line):
     try:
@@ -307,6 +328,18 @@ def check_single_address(ipv, source):
     else:
         return False
 
+def check_mac(mac):
+    if len(mac) == 12+5:
+        # 0 1 : 3 4 : 6 7 : 9 10 : 12 13 : 15 16
+        for i in (2, 5, 8, 11, 14):
+            if mac[i] != ":":
+                return False
+        for i in (0, 1, 3, 4, 6, 7, 9, 10, 12, 13, 15, 16):
+            if mac[i] not in string.hexdigits:
+                return False
+        return True
+    return False
+
 def uniqify(input):
     # removes duplicates from list, whilst preserving order
     output = []
@@ -349,7 +382,7 @@ def checkUid(uid):
     if type(uid) == str:
         try:
             uid = int(uid)
-        except:
+        except ValueError:
             return False
     if uid >= 0 and uid <= 2**31-1:
         return True
