@@ -96,7 +96,7 @@ class FirewallDConfigService(slip.dbus.service.Object):
         elif property_name == "default":
             return self.obj.default
         elif property_name == "builtin":
-            return self.config.is_builtin_service(self.obj)
+            return self.obj.builtin
         else:
             raise dbus.exceptions.DBusException(
                 "org.freedesktop.DBus.Error.AccessDenied: "
@@ -120,6 +120,7 @@ class FirewallDConfigService(slip.dbus.service.Object):
             'filename': self.obj.filename,
             'path': self.obj.path,
             'default': self.obj.default,
+            'builtin': self.obj.builtin,
         }
 
     @slip.dbus.polkit.require_auth(PK_ACTION_CONFIG)
@@ -337,6 +338,57 @@ class FirewallDConfigService(slip.dbus.service.Object):
                    protocol)
         return (port,protocol) in self.getSettings()[3]
 
+    # protocol
+
+    @dbus_service_method(DBUS_INTERFACE_CONFIG_SERVICE, out_signature='as')
+    @dbus_handle_exceptions
+    def getProtocols(self, sender=None):
+        log.debug1("config.service.%d.getProtocols()", self.id)
+        return self.getSettings()[6]
+
+    @dbus_service_method(DBUS_INTERFACE_CONFIG_SERVICE, in_signature='as')
+    @dbus_handle_exceptions
+    def setProtocols(self, protocols, sender=None):
+        protocols = dbus_to_python(protocols, list)
+        log.debug1("config.service.%d.setProtocols('[%s]')", self.id,
+                   ",".join(protocols))
+        self.parent.accessCheck(sender)
+        settings = list(self.getSettings())
+        settings[6] = protocols
+        self.update(settings)
+
+    @dbus_service_method(DBUS_INTERFACE_CONFIG_SERVICE, in_signature='s')
+    @dbus_handle_exceptions
+    def addProtocol(self, protocol, sender=None):
+        protocol = dbus_to_python(protocol, str)
+        log.debug1("config.service.%d.addProtocol('%s')", self.id, protocol)
+        self.parent.accessCheck(sender)
+        settings = list(self.getSettings())
+        if protocol in settings[6]:
+            raise FirewallError(ALREADY_ENABLED, protocol)
+        settings[6].append(protocol)
+        self.update(settings)
+
+    @dbus_service_method(DBUS_INTERFACE_CONFIG_SERVICE, in_signature='s')
+    @dbus_handle_exceptions
+    def removeProtocol(self, protocol, sender=None):
+        protocol = dbus_to_python(protocol, str)
+        log.debug1("config.service.%d.removeProtocol('%s')", self.id, protocol)
+        self.parent.accessCheck(sender)
+        settings = list(self.getSettings())
+        if protocol not in settings[6]:
+            raise FirewallError(NOT_ENABLED, protocol)
+        settings[6].remove(protocol)
+        self.update(settings)
+
+    @dbus_service_method(DBUS_INTERFACE_CONFIG_SERVICE, in_signature='s',
+                         out_signature='b')
+    @dbus_handle_exceptions
+    def queryProtocol(self, protocol, sender=None):
+        protocol = dbus_to_python(protocol, str)
+        log.debug1("config.service.%d.queryProtocol(%s')", self.id, protocol)
+        return protocol in self.getSettings()[6]
+
     # module
 
     @dbus_service_method(DBUS_INTERFACE_CONFIG_SERVICE, out_signature='as')
@@ -429,8 +481,9 @@ class FirewallDConfigService(slip.dbus.service.Object):
                    family, address)
         self.parent.accessCheck(sender)
         settings = list(self.getSettings())
-        if family in settings[5]:
-            raise FirewallError(ALREADY_ENABLED, family)
+        if family in settings[5] and settings[5][family] == address:
+            raise FirewallError(ALREADY_ENABLED, "'%s': '%s'" % (family,
+                                                                 address))
         settings[5][family] = address
         self.update(settings)
 
@@ -444,7 +497,7 @@ class FirewallDConfigService(slip.dbus.service.Object):
         settings = list(self.getSettings())
         if family not in settings[5]:
             raise FirewallError(NOT_ENABLED, family)
-        settings[5].remove(family)
+        del settings[5][family]
         self.update(settings)
 
     @dbus_service_method(DBUS_INTERFACE_CONFIG_SERVICE, in_signature='ss',
