@@ -39,6 +39,7 @@ from firewall.server.config import FirewallDConfig
 from firewall.dbus_utils import dbus_to_python, \
     command_of_sender, context_of_sender, uid_of_sender, user_of_uid
 from firewall.core.io.zone import Zone
+from firewall.core.io.ipset import IPSet
 from firewall.core.io.service import Service
 from firewall.core.io.icmptype import IcmpType
 from firewall.errors import *
@@ -154,6 +155,9 @@ class FirewallD(slip.dbus.service.Object):
         elif prop == "BRIDGE":
             return self.fw.ebtables_enabled
 
+        elif prop == "IPSet":
+            return self.fw.ipset_enabled
+
         else:
             raise dbus.exceptions.DBusException(
                 "org.freedesktop.DBus.Error.AccessDenied: "
@@ -195,6 +199,7 @@ class FirewallD(slip.dbus.service.Object):
             'IPv6': self._get_property("IPv6"),
             'IPv6_rpfilter': self._get_property("IPv6_rpfilter"),
             'BRIDGE': self._get_property("BRIDGE"),
+            'IPSet': self._get_property("IPSet"),
         }
         
 
@@ -777,6 +782,34 @@ class FirewallD(slip.dbus.service.Object):
 
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 
+    # LOG DENIED
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_INFO)
+    @dbus_service_method(DBUS_INTERFACE, in_signature='', out_signature='s')
+    @dbus_handle_exceptions
+    def getLogDenied(self, sender=None):
+        # returns the log denied value
+        log.debug1("getLogDenied()")
+        return self.fw.get_log_denied()
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_CONFIG)
+    @dbus_service_method(DBUS_INTERFACE, in_signature='s', out_signature='')
+    @dbus_handle_exceptions
+    def setLogDenied(self, value, sender=None):
+        # set the log denied value
+        value = dbus_to_python(value, str)
+        log.debug1("setLogDenied('%s')" % value)
+        self.accessCheck(sender)
+        self.fw.set_log_denied(value)
+        self.LogDeniedChanged(value)
+
+    @dbus.service.signal(DBUS_INTERFACE, signature='s')
+    @dbus_handle_exceptions
+    def LogDeniedChanged(self, value):
+        log.debug1("LogDeniedChanged('%s')" % (value))
+
+    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
+
     # DEFAULT ZONE
 
     @slip.dbus.polkit.require_auth(PK_ACTION_INFO)
@@ -1350,6 +1383,88 @@ class FirewallD(slip.dbus.service.Object):
 
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 
+    # PROTOCOLS
+
+    @dbus_handle_exceptions
+    def disableTimedProtocol(self, zone, protocol):
+        log.debug1("zone.disableTimedProtocol('%s', '%s')" % (zone, protocol))
+        del self._timeouts[zone][(protocol)]
+        self.fw.zone.remove_protocol(zone, protocol)
+        self.ProtocolRemoved(zone, protocol)
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_CONFIG)
+    @dbus_service_method(DBUS_INTERFACE_ZONE, in_signature='ssi',
+                         out_signature='s')
+    @dbus_handle_exceptions
+    def addProtocol(self, zone, protocol, timeout, sender=None):
+        # adds protocol <protocol> if not enabled already to zone
+        zone = dbus_to_python(zone, str)
+        protocol = dbus_to_python(protocol, str)
+        timeout = dbus_to_python(timeout, int)
+        log.debug1("zone.enableProtocol('%s', '%s')" % (zone, protocol))
+        self.accessCheck(sender)
+        _zone = self.fw.zone.add_protocol(zone, protocol, timeout, sender)
+
+        if timeout > 0:
+            tag = GLib.timeout_add_seconds(timeout, self.disableTimedProtocol,
+                                           _zone, protocol)
+            self.addTimeout(_zone, protocol, tag)
+
+        self.ProtocolAdded(_zone, protocol, timeout)
+        return _zone
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_CONFIG)
+    @dbus_service_method(DBUS_INTERFACE_ZONE, in_signature='ss',
+                         out_signature='s')
+    @dbus_handle_exceptions
+    def removeProtocol(self, zone, protocol, sender=None):
+        # removes protocol<protocol> if enabled from zone
+        zone = dbus_to_python(zone, str)
+        protocol = dbus_to_python(protocol, str)
+        log.debug1("zone.removeProtocol('%s', '%s')" % (zone, protocol))
+        self.accessCheck(sender)
+        _zone= self.fw.zone.remove_protocol(zone, protocol)
+
+        self.removeTimeout(_zone, protocol)
+        self.ProtocolRemoved(_zone, protocol)
+        return _zone
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_CONFIG_INFO)
+    @dbus_service_method(DBUS_INTERFACE_ZONE, in_signature='ss',
+                         out_signature='b')
+    @dbus_handle_exceptions
+    def queryProtocol(self, zone, protocol, sender=None):
+        # returns true if a protocol is enabled for zone
+        zone = dbus_to_python(zone, str)
+        protocol = dbus_to_python(protocol, str)
+        log.debug1("zone.queryProtocol('%s', '%s')" % (zone, protocol))
+        return self.fw.zone.query_protocol(zone, protocol)
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_CONFIG_INFO)
+    @dbus_service_method(DBUS_INTERFACE_ZONE, in_signature='s',
+                         out_signature='as')
+    @dbus_handle_exceptions
+    def getProtocols(self, zone, sender=None):
+        # returns the list of enabled protocols
+        # TODO: should be renamed to listProtocols()
+        # because is called by firewall-cmd --zone --list-protocols
+        zone = dbus_to_python(zone, str)
+        log.debug1("zone.getProtocols('%s')" % (zone))
+        return self.fw.zone.list_protocols(zone)
+
+    @dbus.service.signal(DBUS_INTERFACE_ZONE, signature='ssi')
+    @dbus_handle_exceptions
+    def ProtocolAdded(self, zone, protocol, timeout=0):
+        log.debug1("zone.ProtocolAdded('%s', '%s', %d)" % \
+                       (zone, protocol, timeout))
+
+    @dbus.service.signal(DBUS_INTERFACE_ZONE, signature='ss')
+    @dbus_handle_exceptions
+    def ProtocolRemoved(self, zone, protocol):
+        log.debug1("zone.ProtocolRemoved('%s', '%s')" % (zone, protocol))
+
+    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
+
     # MASQUERADE
 
     @dbus_handle_exceptions
@@ -1766,14 +1881,14 @@ class FirewallD(slip.dbus.service.Object):
     @dbus.service.signal(DBUS_INTERFACE_DIRECT, signature='sssias')
     @dbus_handle_exceptions
     def RuleAdded(self, ipv, table, chain, priority, args):
-        log.debug1("direct.RuleAdded('%s', '%s', %d, '%s')" % \
-                       (table, chain, priority, "','".join(args)))
+        log.debug1("direct.RuleAdded('%s', '%s', '%s', %d, '%s')" % \
+                       (ipv, table, chain, priority, "','".join(args)))
 
     @dbus.service.signal(DBUS_INTERFACE_DIRECT, signature='sssias')
     @dbus_handle_exceptions
     def RuleRemoved(self, ipv, table, chain, priority, args):
-        log.debug1("direct.RuleRemoved('%s', '%s', %d, '%s')" % \
-                       (table, chain, priority, "','".join(args)))
+        log.debug1("direct.RuleRemoved('%s', '%s', '%s', %d, '%s')" % \
+                       (ipv, table, chain, priority, "','".join(args)))
 
     # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 
@@ -1819,7 +1934,7 @@ class FirewallD(slip.dbus.service.Object):
         self.fw.direct.remove_passthrough(ipv, args)
         self.PassthroughRemoved(ipv, args)
 
-    @slip.dbus.polkit.require_auth(PK_ACTION_DIRECT)
+    @slip.dbus.polkit.require_auth(PK_ACTION_DIRECT_INFO)
     @dbus_service_method(DBUS_INTERFACE_DIRECT, in_signature='sas',
                          out_signature='b')
     @dbus_handle_exceptions
@@ -1831,7 +1946,7 @@ class FirewallD(slip.dbus.service.Object):
                        (ipv, "','".join(args)))
         return self.fw.direct.query_passthrough(ipv, args)
 
-    @slip.dbus.polkit.require_auth(PK_ACTION_DIRECT)
+    @slip.dbus.polkit.require_auth(PK_ACTION_DIRECT_INFO)
     @dbus_service_method(DBUS_INTERFACE_DIRECT, in_signature='',
                          out_signature='a(sas)')
     @dbus_handle_exceptions
@@ -1851,7 +1966,7 @@ class FirewallD(slip.dbus.service.Object):
         for passthrough in reversed(self.getAllPassthroughs()):
             self.removePassthrough(*passthrough)
 
-    @slip.dbus.polkit.require_auth(PK_ACTION_DIRECT)
+    @slip.dbus.polkit.require_auth(PK_ACTION_DIRECT_INFO)
     @dbus_service_method(DBUS_INTERFACE_DIRECT, in_signature='s',
                          out_signature='aas')
     @dbus_handle_exceptions
@@ -1884,3 +1999,97 @@ class FirewallD(slip.dbus.service.Object):
             Use-case is GUI (RHBZ#994729).
         """
         pass
+
+    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
+    # IPSETS
+    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_INFO)
+    @dbus_service_method(DBUS_INTERFACE_IPSET, in_signature='s',
+                         out_signature='b')
+    @dbus_handle_exceptions
+    def queryIPSet(self, ipset, sender=None):
+        # returns true if a set with the name exists
+        ipset = dbus_to_python(ipset)
+        log.debug1("ipset.queryIPSet('%s')" % (ipset))
+        return self.fw.ipset.query_ipset(ipset)
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_INFO)
+    @dbus_service_method(DBUS_INTERFACE_IPSET, in_signature='',
+                         out_signature='as')
+    @dbus_handle_exceptions
+    def getIPSets(self, sender=None):
+        # returns list of added sets
+        log.debug1("ipsets.getIPSets()")
+        return self.fw.ipset.get_ipsets()
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_INFO)
+    @dbus_service_method(DBUS_INTERFACE_IPSET, in_signature='s',
+                         out_signature=IPSet.DBUS_SIGNATURE)
+    @dbus_handle_exceptions
+    def getIPSetSettings(self, ipset, sender=None):
+        # returns ipset settings for ipset
+        ipset = dbus_to_python(ipset, str)
+        log.debug1("getIPSetSettings(%s)", ipset)
+        return self.fw.ipset.get_ipset(ipset).export_config()
+
+    # set entries # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
+
+    @dbus_service_method(DBUS_INTERFACE_IPSET, in_signature='ss',
+                         out_signature='')
+    @dbus_handle_exceptions
+    def addEntry(self, ipset, entry, sender=None):
+        # adds ipset entry
+        ipset = dbus_to_python(ipset)
+        entry = dbus_to_python(entry)
+        log.debug1("ipset.addEntry('%s', '%s')" % (ipset, entry))
+        self.accessCheck(sender)
+        self.fw.ipset.add_entry(ipset, entry)
+        self.EntryAdded(ipset, entry)
+
+    @dbus_service_method(DBUS_INTERFACE_IPSET, in_signature='ss',
+                         out_signature='')
+    @dbus_handle_exceptions
+    def removeEntry(self, ipset, entry, sender=None):
+        # removes ipset entry
+        ipset = dbus_to_python(ipset)
+        entry = dbus_to_python(entry)
+        log.debug1("ipset.removeEntry('%s', '%s')" % (ipset, entry))
+        self.accessCheck(sender)
+        self.fw.ipset.remove_entry(ipset, entry)
+        self.EntryRemoved(ipset, entry)
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_INFO)
+    @dbus_service_method(DBUS_INTERFACE_IPSET, in_signature='ss',
+                         out_signature='b')
+    @dbus_handle_exceptions
+    def queryEntry(self, ipset, entry, sender=None):
+        # returns true if the entry exists in the ipset
+        ipset = dbus_to_python(ipset)
+        entry = dbus_to_python(entry)
+        log.debug1("ipset.queryEntry('%s', '%s')" % (ipset, entry))
+        return self.fw.ipset.query_entry(ipset, entry)
+
+    @slip.dbus.polkit.require_auth(PK_ACTION_INFO)
+    @dbus_service_method(DBUS_INTERFACE_IPSET, in_signature='s',
+                         out_signature='as')
+    @dbus_handle_exceptions
+    def getEntries(self, ipset, sender=None):
+        # returns list of added entries for the ipset
+        ipset = dbus_to_python(ipset)
+        log.debug1("ipset.getEntries('%s')" % ipset)
+        return self.fw.ipset.get_entries(ipset)
+
+    @dbus.service.signal(DBUS_INTERFACE_IPSET, signature='ss')
+    @dbus_handle_exceptions
+    def EntryAdded(self, ipset, entry):
+        ipset = dbus_to_python(ipset)
+        entry = dbus_to_python(entry)
+        log.debug1("ipset.EntryAdded('%s', '%s')" % (ipset, entry))
+
+    @dbus.service.signal(DBUS_INTERFACE_IPSET, signature='ss')
+    @dbus_handle_exceptions
+    def EntryRemoved(self, ipset, entry):
+        ipset = dbus_to_python(ipset)
+        entry = dbus_to_python(entry)
+        log.debug1("ipset.EntryRemoved('%s', '%s')" % (ipset, entry))
